# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
Implementation of the XEP-0045: Multi-User Chat.
Add some facilities that are not available on the XEP_0045
slix plugin
"""

from __future__ import annotations

from typing import (
    cast,
    Optional,
    TYPE_CHECKING,
)

from slixmpp import (
    JID,
    ClientXMPP,
    Presence,
)
from slixmpp.exceptions import IqError, IqTimeout
from slixmpp.types import PresenceArgs, PresenceShows

import logging
log = logging.getLogger(__name__)


if TYPE_CHECKING:
    from poezio.core.core import Core
    from poezio.tabs import MucTab


def change_show(
    xmpp: ClientXMPP,
    jid: JID,
    own_nick: str,
    show: str | None,
    status: Optional[str]
) -> None:
    """
    Change our 'Show'
    """
    jid = JID(jid)
    pres: Presence = xmpp.make_presence(pto='%s/%s' % (jid, own_nick))
    if show:  # if show is None, don't put a <show /> tag. It means "available"
        pres['type'] = show
    if status:
        pres['status'] = status
    pres.send()


def change_nick(
    core: Core,
    jid: JID,
    nick: str,
    status: Optional[str] = None,
    show: Optional[str] = None
) -> None:
    """
    Change our own nick in a room
    """
    xmpp = core.xmpp
    pargs = PresenceArgs(
        pshow=cast(PresenceShows, show),
    )
    if status:
        pargs['pstatus'] = status
    core.register_task(xmpp.plugin['xep_0045'].set_self_nick(
        JID(jid),
        nick,
        presence_options=pargs
    ))
    core.events.trigger('changing_nick', jid, nick)


def join_groupchat(
    core: Core,
    jid: JID,
    nick: str,
    passwd: str = '',
    status: Optional[str] = None,
    show: Optional[str] = None,
    seconds: Optional[int] = None,
    tab: Optional['MucTab'] = None
) -> None:
    xmpp = core.xmpp

    async def disco_and_join() -> None:
        has_mam, has_occupant_id, has_moderation = False, False, False
        has_stanzaid = False
        try:
            iq = await xmpp.plugin['xep_0030'].get_info(jid=jid)
            features = iq['disco_info'].get_features()
            has_mam = 'urn:xmpp:mam:2' in features
            has_occupant_id = 'urn:xmpp:occupant-id:0' in features
            has_moderation = 'urn:xmpp:message-moderate:1' in features
            has_stanzaid = 'urn:xmpp:sid:0' in features
        except (IqError, IqTimeout):
            pass
        secs: int | None
        if tab:
            tab.supports_occupantid = has_occupant_id
            tab.supports_moderation = has_moderation
            tab.supports_stanzaid = has_stanzaid
        if has_mam or (tab and tab._text_buffer.last_message):
            secs = 0
        else:
            secs = seconds
        core.events.trigger('joining_muc', jid)
        pargs = PresenceArgs()
        if status:
            pargs['pstatus'] = status
        if show:
            pargs['pshow'] = cast(PresenceShows, show)
        await xmpp.plugin['xep_0045'].join_muc_wait(
            room=jid,
            nick=nick,
            seconds=secs,
            password=passwd,
            presence_options=pargs,
        )

    core.register_task(disco_and_join())


def leave_groupchat(
    xmpp: ClientXMPP,
    jid: JID,
    own_nick: str,
    msg: str
) -> None:
    """
    Leave the groupchat
    """
    jid = JID(jid)
    try:
        xmpp.plugin['xep_0045'].leave_muc(jid, own_nick, msg)
    except KeyError:
        log.debug(
            "muc.leave_groupchat: could not leave the room %s",
            jid,
            exc_info=True)



"""
Repeats the last word of the last message in the conversation, and use it in
an annoying “C’est toi le” sentence.

Installation
------------

You only have to load the plugin:

.. code-block:: none

    /load stoi

.. glossary::

    /stoi
        **Usage:** ``/stoi``

"""
from poezio.plugin import BasePlugin
from poezio import tabs
import string
from poezio import xhtml
import random

char_we_dont_want = string.punctuation + ' ’„“”…«»'


class Plugin(BasePlugin):
    def init(self) -> None:
        for tab_type in (tabs.MucTab, tabs.PrivateTab, tabs.DynamicConversationTab, tabs.StaticConversationTab):
            self.api.add_tab_command(
                tab_type,
                'stoi',
                handler=self.stoi,
                help="Repeats the last word of the last "
                "message in the conversation, and "
                "use it in an annoying “C’est toi "
                "le” sentence.",
                short='C’est toi le stoi.')

    def stoi(self, args: str) -> None:
        messages = self.api.get_conversation_messages()
        if not messages:
            # Do nothing if the conversation doesn’t contain any message
            return
        last_message = messages[-1]
        txt = xhtml.clean_text(last_message.txt)
        for char in char_we_dont_want:
            txt = txt.replace(char, ' ')
        if txt.strip():
            last_word = txt.split()[-1]
        else:
            last_word = "vide"
        intro = "C'est toi " if random.getrandbits(1) else "Stoi "
        if last_word[0] in 'aeiouAEIOUÀàÉéÈè':
            msg = intro + ('l’%s' % last_word)
        else:
            msg = intro + ('le %s' % last_word)
        self.api.send_message(msg)



"""
Windows used by the DataFormsTab.

We only need to export the FormWin (which is not a real Win, as it
does not inherit from the Win base class), as it will create the
others when needed.
"""

from poezio.windows import base_wins
from poezio.windows.base_wins import Win
from poezio.windows.inputs import Input

from poezio.theming import to_curses_attr, get_theme, TextStyle
from slixmpp.plugins.xep_0004.stanza import Form, FormField

from typing import TypedDict, Any


class InputDict(TypedDict):
    label: 'ColoredLabel'
    description: str
    input: 'FieldTypes'


class FieldInput:
    """
    All input types in a data form should inherit this class,
    in addition with windows.Input or any relevant class from the
    'windows' library.
    """
    def is_dummy(self) -> bool:
        return False

    def reply(self) -> None:
        """
        Set the correct response value in the field
        """
        raise NotImplementedError

    def get_help_message(self) -> str:
        """
        Should return a string explaining the keys of the input.
        Will be displayed at each refresh on a line at the bottom of the tab.
        """
        return ''


class FieldInputWin(Win):
    "Mix both FieldInput and Win"

    __slots__ = ('color',)
    color: TextStyle

    def __init__(self) -> None:
        Win.__init__(self)
        self.color = get_theme().COLOR_NORMAL_TEXT

    def is_dummy(self) -> bool:
        return False

    def set_color(self, color: TextStyle) -> None:
        self.color = color
        self.refresh()

    def do_command(self, key: str, reset: bool = True, raw: bool = False) -> bool:
        raise NotImplementedError()

    def refresh(self) -> None:
        raise NotImplementedError()

    def reply(self) -> None:
        """
        Set the correct response value in the field
        """
        raise NotImplementedError

    def get_help_message(self) -> str:
        """
        Should return a string explaining the keys of the input.
        Will be displayed at each refresh on a line at the bottom of the tab.
        """
        return ''


class DataFormField(FieldInput):
    """Specialize the field for data forms"""
    _field: FormField

    def __init__(self, field: FormField) -> None:
        super().__init__()
        self._field = field


class DataFormFieldWin(FieldInputWin):
    """Specialize the field for data forms"""
    _field: FormField

    __slots__ = ('_field',)

    def __init__(self, field: FormField) -> None:
        super().__init__()
        self._field = field


class ColoredLabel(Win):
    __slots__ = ('text', 'color')

    color: TextStyle

    def __init__(self, text: str) -> None:
        self.text = text
        self.color = get_theme().COLOR_NORMAL_TEXT
        Win.__init__(self)

    def set_color(self, color: TextStyle) -> None:
        self.color = color
        self.refresh()

    def refresh(self) -> None:
        self._win.erase()
        self._win.attron(to_curses_attr(self.color))
        self.addstr(0, 0, self.text)
        self._win.attroff(to_curses_attr(self.color))
        self._refresh()


class DummyInput(DataFormFieldWin):
    """
    Used for fields that do not require any input ('fixed')
    """

    __slots__ = ()

    def __init__(self, field: FormField) -> None:
        DataFormFieldWin.__init__(self, field)

    def do_command(self, *args: Any, **kwargs: Any) -> bool:
        return False

    def refresh(self) -> None:
        return

    def is_dummy(self) -> bool:
        return True


class BooleanWin(DataFormFieldWin):
    __slots__ = ('last_key', 'value')

    def __init__(self, field: FormField) -> None:
        DataFormFieldWin.__init__(self, field)
        self.last_key = 'KEY_RIGHT'
        self.value = bool(field.get_value())

    def do_command(self, key: str, reset: bool = True, raw: bool = False) -> bool:
        if key == 'KEY_LEFT' or key == 'KEY_RIGHT':
            self.value = not self.value
            self.last_key = key
        self.refresh()
        return True

    def refresh(self) -> None:
        self._win.erase()
        self._win.attron(to_curses_attr(self.color))
        self.addnstr(0, 0, ' ' * (8), self.width)
        self.addstr(0, 2, "%s" % self.value)
        self.addstr(0, 8, '→')
        self.addstr(0, 0, '←')
        if self.last_key == 'KEY_RIGHT':
            self.addstr(0, 8, '')
        else:
            self.addstr(0, 0, '')
        self._win.attroff(to_curses_attr(self.color))
        self._refresh()

    def reply(self) -> None:
        self._field['label'] = ''
        self._field.set_answer(self.value)

    def get_help_message(self) -> str:
        return '← and →: change the value between True and False'


class TextMultiWin(DataFormFieldWin):
    __slots__ = ('options', 'val_pos', 'edition_input')
    edition_input: 'Input | None'

    def __init__(self, field: FormField) -> None:
        DataFormFieldWin.__init__(self, field)
        options = field.get_value()
        if isinstance(options, list):
            self.options = options
        else:
            self.options = options.split('\n') if options else []
        self.val_pos = 0
        self.edition_input = None
        if not isinstance(self.options, list):
            if isinstance(self.options, str):
                self.options = [self.options]
            else:
                self.options = []
        self.options.append('')

    def do_command(self, key: str, reset: bool = True, raw: bool = False) -> bool:
        if not self.edition_input:
            if key == 'KEY_LEFT':
                if self.val_pos > 0:
                    self.val_pos -= 1
            elif key == 'KEY_RIGHT':
                if self.val_pos < len(self.options) - 1:
                    self.val_pos += 1
            elif key == '^M':
                edition_input = Input()
                self.edition_input = edition_input
                edition_input.color = self.color
                edition_input.resize(self.height, self.width, self.y,
                                     self.x)
                edition_input.text = self.options[self.val_pos]
                edition_input.key_end()
        else:
            if key == '^M':
                self.options[self.val_pos] = self.edition_input.get_text()
                if not self.options[self.val_pos] and self.val_pos != len(
                        self.options) - 1:
                    del self.options[self.val_pos]
                    if self.val_pos == len(self.options) - 1:
                        self.val_pos -= 1
                self.edition_input = None
                if not self.options or self.options[-1] != '':
                    self.options.append('')
            else:
                self.edition_input.do_command(key, raw=raw)
        self.refresh()
        return True

    def refresh(self) -> None:
        if not self.edition_input:
            self._win.erase()
            self._win.attron(to_curses_attr(self.color))
            self.addnstr(0, 0, ' ' * self.width, self.width)
            option = self.options[self.val_pos]
            self.addstr(0, self.width // 2 - len(option) // 2, option)
            if self.val_pos > 0:
                self.addstr(0, 0, '←')
            if self.val_pos < len(self.options) - 1:
                self.addstr(0, self.width - 1, '→')
            self._win.attroff(to_curses_attr(self.color))
            self._refresh()
        else:
            self.edition_input.refresh()

    def reply(self) -> None:
        values = [val for val in self.options if val]
        self._field.set_answer(values)

    def get_help_message(self) -> str:
        if not self.edition_input:
            help_msg = '← and →: browse the available entries. '
            if self.val_pos == len(self.options) - 1:
                help_msg += 'Enter: add an entry'
            else:
                help_msg += 'Enter: edit this entry'
        else:
            help_msg = 'Enter: finish editing this entry.'
        return help_msg


class ListMultiWin(DataFormFieldWin):
    __slots__ = ('options', 'val_pos')

    def __init__(self, field: FormField) -> None:
        DataFormFieldWin.__init__(self, field)
        values = field.get_value() or []
        self.options = [[option, True if option['value'] in values else False]
                        for option in field.get_options()]
        self.val_pos = 0

    def do_command(self, key: str, reset: bool = True, raw: bool = False) -> bool:
        if key == 'KEY_LEFT':
            if self.val_pos > 0:
                self.val_pos -= 1
        elif key == 'KEY_RIGHT':
            if self.val_pos < len(self.options) - 1:
                self.val_pos += 1
        elif key == ' ':
            self.options[self.val_pos][1] = not self.options[self.val_pos][1]
        else:
            return True
        self.refresh()
        return True

    def refresh(self) -> None:
        self._win.erase()
        self._win.attron(to_curses_attr(self.color))
        self.addnstr(0, 0, ' ' * self.width, self.width)
        if self.val_pos > 0:
            self.addstr(0, 0, '←')
        if self.val_pos < len(self.options) - 1:
            self.addstr(0, self.width - 1, '→')
        if self.options:
            option = self.options[self.val_pos]
            self.addstr(0, self.width // 2 - len(option) // 2,
                        option[0]['label'])
            self.addstr(0, 2, '✔' if option[1] else '☐')
        self._win.attroff(to_curses_attr(self.color))
        self._refresh()

    def reply(self) -> None:
        self._field['label'] = ''
        self._field.del_options()
        values = [
            option[0]['value'] for option in self.options if option[1] is True
        ]
        self._field.set_answer(values)

    def get_help_message(self) -> str:
        return '←, →: Switch between the value. Space: select or deselect a value'


class ListSingleWin(DataFormFieldWin):
    __slots__ = ('options', 'val_pos')

    def __init__(self, field: FormField) -> None:
        DataFormFieldWin.__init__(self, field)
        # the option list never changes
        self.options = field.get_options()
        # val_pos is the position of the currently selected option
        self.val_pos = 0
        for i, option in enumerate(self.options):
            if field.get_value() == option['value']:
                self.val_pos = i

    def do_command(self, key: str, reset: bool = True, raw: bool = False) -> bool:
        if key == 'KEY_LEFT':
            if self.val_pos > 0:
                self.val_pos -= 1
        elif key == 'KEY_RIGHT':
            if self.val_pos < len(self.options) - 1:
                self.val_pos += 1
        else:
            return True
        self.refresh()
        return True

    def refresh(self) -> None:
        self._win.erase()
        self._win.attron(to_curses_attr(self.color))
        self.addnstr(0, 0, ' ' * self.width, self.width)
        if self.val_pos > 0:
            self.addstr(0, 0, '←')
        if self.val_pos < len(self.options) - 1:
            self.addstr(0, self.width - 1, '→')
        if self.options:
            option = self.options[self.val_pos]['label']
            self.addstr(0, self.width // 2 - len(option) // 2, option)
        self._win.attroff(to_curses_attr(self.color))
        self._refresh()

    def reply(self) -> None:
        self._field['label'] = ''
        self._field.del_options()
        self._field.set_answer(self.options[self.val_pos]['value'])

    def get_help_message(self) -> str:
        return '←, →: Select a value amongst the others'


class TextSingleWin(DataFormField, Input):
    __slots__ = ('text', 'pos')

    def __init__(self, field: FormField) -> None:
        DataFormField.__init__(self, field)
        Input.__init__(self)
        self.text = field.get_value() if isinstance(field.get_value(), str)\
            else ""
        self.pos = 0
        self.view_pos = 0
        self.color = get_theme().COLOR_NORMAL_TEXT

    def reply(self) -> None:
        self._field['label'] = ''
        self._field.set_answer(self.get_text())

    def get_help_message(self) -> str:
        return 'Edit the text'


class TextPrivateWin(TextSingleWin):
    __slots__ = ()

    def __init__(self, field: FormField) -> None:
        TextSingleWin.__init__(self, field)

    def rewrite_text(self) -> None:
        self._win.erase()
        if self.color:
            self._win.attron(to_curses_attr(self.color))
        self.addstr(
            '*' * len(self.text[self.view_pos:self.view_pos + self.width - 1]))
        if self.color:
            (y, x) = self._win.getyx()
            size = self.width - x
            self.addnstr(' ' * size, size, to_curses_attr(self.color))
        self.addstr(0, self.pos, '')
        if self.color:
            self._win.attroff(to_curses_attr(self.color))
        self._refresh()

    def get_help_message(self) -> str:
        return 'Edit the secret text'


FieldTypes = DummyInput | BooleanWin | TextMultiWin | ListMultiWin | ListSingleWin | TextSingleWin | TextPrivateWin


class FormWin:
    """
    A window, with some subwins (the various inputs).
    On init, create all the subwins.
    On resize, move and resize all the subwin and define how the text will be written
    On refresh, write all the text, and refresh all the subwins
    """
    __slots__ = ('_win', 'height', 'width', '_form', 'scroll_pos', 'current_input', 'inputs')

    input_classes: dict[str, type[FieldTypes]] = {
        'boolean': BooleanWin,
        'fixed': DummyInput,
        'jid-multi': TextMultiWin,
        'jid-single': TextSingleWin,
        'list-multi': ListMultiWin,
        'list-single': ListSingleWin,
        'text-multi': TextMultiWin,
        'text-private': TextPrivateWin,
        'text-single': TextSingleWin,
    }

    def __init__(self, form: Form, height: int, width: int, y: int, x: int) -> None:
        self._form = form
        if base_wins.TAB_WIN:
            win = base_wins.TAB_WIN.derwin(height, width, y, x)
            if not win:
                raise ValueError()
            self._win = win
        else:
            raise ValueError()
        self.scroll_pos = 0
        self.current_input = 0
        self.inputs: list[InputDict] = []  # dict list
        for (_name, field) in self._form.get_fields().items():
            if field['type'] == 'hidden':
                continue
            input_class: type[FieldTypes]
            if field['type'] not in self.input_classes:
                input_class = TextSingleWin
            else:
                input_class = self.input_classes[field['type']]
            label = field['label']
            desc = field['desc']
            if field['type'] == 'fixed':
                label = field.get_value()
            inp = input_class(field)
            self.inputs.append({
                'label': ColoredLabel(label),
                'description': desc,
                'input': inp
            })

    def resize(self, height: int, width: int, y: int, x: int) -> None:
        self.height = height
        self.width = width
        if base_wins.TAB_WIN:
            self._win = base_wins.TAB_WIN.derwin(height, width, y, x)
        # Adjust the scroll position, if resizing made the window too small
        # for the cursor to be visible
        while self.current_input - self.scroll_pos > self.height - 1:
            self.scroll_pos += 1

    def reply(self) -> None:
        """
        Set the response values in the form, for each field
        from the corresponding input
        """
        for inp in self.inputs:
            if inp['input'].is_dummy():
                continue
            else:
                inp['input'].reply()
        self._form['title'] = ''
        self._form['instructions'] = ''

    def go_to_next_input(self) -> None:
        if not self.inputs:
            return
        if self.current_input == len(self.inputs) - 1:
            return
        theme = get_theme()
        self.inputs[self.current_input]['input'].set_color(
            theme.COLOR_NORMAL_TEXT)
        self.inputs[self.current_input]['label'].set_color(
            theme.COLOR_NORMAL_TEXT)
        self.current_input += 1
        jump = 0
        while self.current_input + jump != len(
                self.inputs) - 1 and self.inputs[self.current_input +
                                                 jump]['input'].is_dummy():
            jump += 1
        if self.inputs[self.current_input + jump]['input'].is_dummy():
            return
        self.current_input += jump
        # If moving made the current input out of the visible screen, we
        # adjust the scroll position and we redraw the whole thing. We don’t
        # call refresh() if this is not the case, because
        # refresh_current_input() is always called anyway, so this is not
        # needed
        if self.current_input - self.scroll_pos > self.height - 1:
            self.scroll_pos += 1
            self.refresh()
        self.inputs[self.current_input]['input'].set_color(
            theme.COLOR_SELECTED_ROW)
        self.inputs[self.current_input]['label'].set_color(
            theme.COLOR_SELECTED_ROW)

    def go_to_previous_input(self) -> None:
        if not self.inputs:
            return
        if self.current_input == 0:
            return
        theme = get_theme()
        self.inputs[self.current_input]['input'].set_color(
            theme.COLOR_NORMAL_TEXT)
        self.inputs[self.current_input]['label'].set_color(
            theme.COLOR_NORMAL_TEXT)
        self.current_input -= 1
        jump = 0
        while self.current_input - jump > 0 and self.inputs[self.current_input
                                                            +
                                                            jump]['input'].is_dummy(
                                                            ):
            jump += 1
        if self.inputs[self.current_input + jump]['input'].is_dummy():
            return
        # Adjust the scroll position if the current_input would be outside
        # of the visible area
        if self.current_input < self.scroll_pos:
            self.scroll_pos = self.current_input
            self.refresh()
        self.current_input -= jump
        self.inputs[self.current_input]['input'].set_color(
            theme.COLOR_SELECTED_ROW)
        self.inputs[self.current_input]['label'].set_color(
            theme.COLOR_SELECTED_ROW)

    def on_input(self, key: str, raw: bool = False) -> None:
        if not self.inputs:
            return
        self.inputs[self.current_input]['input'].do_command(key, raw=raw)

    def refresh(self) -> None:
        self._win.erase()
        y = -self.scroll_pos
        i = 0
        for _name, field in self._form.get_fields().items():
            if field['type'] == 'hidden':
                continue
            self.inputs[i]['label'].resize(1, self.width // 2, y + 1, 0)
            self.inputs[i]['input'].resize(1, self.width // 2, y + 1,
                                           self.width // 2)
            # TODO: display the field description
            y += 1
            i += 1
        self._win.refresh()
        for i, inp in enumerate(self.inputs):
            if i < self.scroll_pos:
                continue
            if i >= self.height + self.scroll_pos:
                break
            inp['label'].refresh()
            inp['input'].refresh()
            inp['label'].refresh()
        if self.inputs and self.current_input < self.height - 1:
            color = get_theme().COLOR_SELECTED_ROW
            self.inputs[self.current_input]['input'].set_color(color)
            self.inputs[self.current_input]['input'].refresh()
            self.inputs[self.current_input]['label'].set_color(color)
            self.inputs[self.current_input]['label'].refresh()

    def refresh_current_input(self) -> None:
        self.inputs[self.current_input]['input'].refresh()

    def get_help_message(self) -> str:
        if (self.inputs and self.current_input < self.height - 1
                and self.inputs[self.current_input]['input']):
            return self.inputs[self.current_input]['input'].get_help_message()
        return ''



from poezio.tabs.basetabs import (
    Tab, ChatTab, GapTab, OneToOneTab, DefaultHelpTab, NoInputTab,
)
from poezio.tabs.basetabs import STATE_PRIORITY, SHOW_NAME
from poezio.tabs.rostertab import RosterInfoTab
from poezio.tabs.muctab import MucTab, NS_MUC_USER
from poezio.tabs.privatetab import PrivateTab
from poezio.tabs.confirmtab import ConfirmTab
from poezio.tabs.conversationtab import (
    ConversationTab,
    StaticConversationTab,
    DynamicConversationTab,
)
from poezio.tabs.xmltab import XMLTab
from poezio.tabs.listtab import ListTab
from poezio.tabs.muclisttab import MucListTab
from poezio.tabs.adhoc_commands_list import AdhocCommandsListTab
from poezio.tabs.data_forms import DataFormsTab
from poezio.tabs.bookmarkstab import BookmarksTab

__all__ = [
    'Tab', 'ChatTab', 'GapTab', 'OneToOneTab', 'STATE_PRIORITY', 'SHOW_NAME',
    'RosterInfoTab', 'MucTab', 'NS_MUC_USER', 'PrivateTab', 'ConfirmTab',
    'ConversationTab', 'StaticConversationTab', 'DynamicConversationTab',
    'XMLTab', 'ListTab', 'MucListTab', 'AdhocCommandsListTab', 'DataFormsTab',
    'BookmarksTab', 'DefaultHelpTab', 'NoInputTab',
]



"""
Test the functions in the `theming` module
"""

import pytest

from poezio.theming import dump_tuple, read_tuple

def test_read_tuple():
    assert read_tuple('1,-1,u') == ((1, -1), 'u')
    assert read_tuple('1,2') == ((1, 2), '\0')

    with pytest.raises(IndexError):
        read_tuple('1')

    with pytest.raises(ValueError):
        read_tuple('toto')

def test_dump_tuple():
    assert dump_tuple((1, 2)) == '1,2'
    assert dump_tuple((1, )) == '1'
    assert dump_tuple((1, 2, 'u')) == '1,2,u'





"""
Module for the base Tabs

The root class Tab defines the generic interface and attributes of a
tab. A tab organizes various Windows around the screen depending
of the tab specificity. If the tab shows messages, it will also
reference a buffer containing the messages.

Each subclass should redefine its own refresh() and resize() method
according to its windows.

This module also defines ChatTabs, the parent class for all tabs
revolving around chats.
"""

from __future__ import annotations

import logging
import string
import asyncio
import curses
from copy import copy
from math import ceil, log10
from datetime import datetime
from inspect import iscoroutine
from xml.etree import ElementTree as ET
from xml.sax import SAXParseException
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Optional,
    Union,
    TYPE_CHECKING,
)

from poezio import (
    poopt,
    timed_events,
    xhtml,
    windows
)
from poezio.core.structs import Command, Completion, Status
from poezio.config import config
from poezio.decorators import command_args_parser, refresh_wrapper
from poezio.logger import logger
from poezio.log_loader import MAMFiller, LogLoader
from poezio.text_buffer import TextBuffer, DeletionError
from poezio.theming import get_theme, dump_tuple, TextStyle
from poezio.user import User
from poezio.ui.funcs import truncate_nick
from poezio.timed_events import DelayedEvent
from poezio.ui.types import (
    BaseMessage,
    Message,
    PersistentInfoMessage,
    LoggableTrait,
)

from slixmpp import JID, InvalidJID, Message as SMessage

if TYPE_CHECKING:
    from poezio.size_manager import SizeManager
    from poezio.core.core import Core
    from poezio.windows import Input

log = logging.getLogger(__name__)

# getters for tab colors (lambdas, so that they are dynamic)
STATE_COLORS = {
    'disconnected': lambda: get_theme().COLOR_TAB_DISCONNECTED,
    'scrolled': lambda: get_theme().COLOR_TAB_SCROLLED,
    'nonempty': lambda: get_theme().COLOR_TAB_NONEMPTY,
    'joined': lambda: get_theme().COLOR_TAB_JOINED,
    'message': lambda: get_theme().COLOR_TAB_NEW_MESSAGE,
    'composing': lambda: get_theme().COLOR_TAB_COMPOSING,
    'highlight': lambda: get_theme().COLOR_TAB_HIGHLIGHT,
    'private': lambda: get_theme().COLOR_TAB_PRIVATE,
    'normal': lambda: get_theme().COLOR_TAB_NORMAL,
    'current': lambda: get_theme().COLOR_TAB_CURRENT,
    'attention': lambda: get_theme().COLOR_TAB_ATTENTION,
}
VERTICAL_STATE_COLORS = {
    'disconnected': lambda: get_theme().COLOR_VERTICAL_TAB_DISCONNECTED,
    'scrolled': lambda: get_theme().COLOR_VERTICAL_TAB_SCROLLED,
    'nonempty': lambda: get_theme().COLOR_VERTICAL_TAB_NONEMPTY,
    'joined': lambda: get_theme().COLOR_VERTICAL_TAB_JOINED,
    'message': lambda: get_theme().COLOR_VERTICAL_TAB_NEW_MESSAGE,
    'composing': lambda: get_theme().COLOR_VERTICAL_TAB_COMPOSING,
    'highlight': lambda: get_theme().COLOR_VERTICAL_TAB_HIGHLIGHT,
    'private': lambda: get_theme().COLOR_VERTICAL_TAB_PRIVATE,
    'normal': lambda: get_theme().COLOR_VERTICAL_TAB_NORMAL,
    'current': lambda: get_theme().COLOR_VERTICAL_TAB_CURRENT,
    'attention': lambda: get_theme().COLOR_VERTICAL_TAB_ATTENTION,
}

# priority of the different tab states when using Alt+e
# higher means more priority, < 0 means not selectable
STATE_PRIORITY = {
    'normal': -1,
    'current': -1,
    'disconnected': 0,
    'nonempty': 0.1,
    'scrolled': 0.5,
    'joined': 0.8,
    'composing': 0.9,
    'message': 1,
    'highlight': 2,
    'private': 2,
    'attention': 3
}

SHOW_NAME = {
    'dnd': 'busy',
    'away': 'away',
    'xa': 'not available',
    'chat': 'chatty',
    '': 'available'
}


class Tab:
    plugin_commands: Dict[str, Command] = {}
    plugin_keys: Dict[str, Callable[..., Any]] = {}
    # Placeholder values, set on resize
    height: ClassVar[int] = 1
    width: ClassVar[int] = 1
    core: Core
    input: windows.InputLike
    key_func: Dict[str, Callable[[], Any]]
    commands: Dict[str, Command]
    need_resize: bool
    ui_config_changed: bool
    _prev_state: str | None

    def __init__(self, core: Core):
        self.core = core
        self.nb = 0
        if not hasattr(self, '_name'):
            self._name = self.__class__.__name__
        self.input = self._create_initial_input()
        self.closed = False
        self._state = 'normal'
        self._prev_state = None

        self.need_resize = False
        self.ui_config_changed = False
        self.key_func = {}  # each tab should add their keys in there
        # and use them in on_input
        self.commands = {}  # and their own commands

    def _create_initial_input(self) -> windows.InputLike:
        """To reimplement in child classes to create the input"""
        raise NotImplementedError()

    @property
    def name(self) -> str:
        if hasattr(self, '_name'):
            return self._name
        return ''

    @property
    def size(self) -> SizeManager:
        return self.core.size

    @staticmethod
    def tab_win_height() -> int:
        """
        Returns 1 or 0, depending on if we are using the vertical tab list
        or not.
        """
        if config.getbool('enable_vertical_tab_list'):
            return 0
        return 1

    @property
    def info_win(self) -> windows.TextWin:
        return self.core.information_win

    @property
    def color(self) -> TextStyle:
        return STATE_COLORS[self._state]()

    @property
    def vertical_color(self) -> TextStyle:
        return VERTICAL_STATE_COLORS[self._state]()

    @property
    def priority(self) -> Union[int, float]:
        return STATE_PRIORITY.get(self._state, -1)

    @property
    def state(self) -> str:
        return self._state

    @state.setter
    def state(self, value: str) -> None:
        if value not in STATE_COLORS:
            log.debug("Invalid value for tab state: %s", value)
        elif STATE_PRIORITY[value] < STATE_PRIORITY[self._state] and \
                value not in ('current', 'disconnected') and \
                not (self._state == 'scrolled' and value == 'disconnected'):
            log.debug(
                "Did not set state because of lower priority, asked: %s, kept: %s",
                value, self._state)
        elif self._state == 'disconnected' and value not in ('joined',
                                                             'current'):
            log.debug(
                'Did not set state because disconnected tabs remain visible')
        else:
            self._state = value
            if self._state == 'current':
                self._prev_state = None

    def set_state(self, value: str) -> None:
        self._state = value

    def save_state(self) -> None:
        if self._state != 'composing':
            self._prev_state = self._state

    def restore_state(self) -> None:
        if self.state == 'composing' and self._prev_state:
            self._state = self._prev_state
            self._prev_state = None
        elif not self._prev_state:
            self._state = 'normal'

    @staticmethod
    def initial_resize(scr: curses.window) -> None:
        Tab.height, Tab.width = scr.getmaxyx()
        windows.base_wins.TAB_WIN = scr

    def information(self, msg: str, typ: str = '') -> bool:
        """Shorthand for self.core.information"""
        return self.core.information(msg, typ)

    def missing_command_callback(self, command_name: str) -> bool:
        """
        Callback executed when a command is not found.
        Returns True if the callback took care of displaying
        the error message, False otherwise.
        """
        return False

    def register_commands_batch(self, commands: list[dict[str, Any]]) -> None:
        """
        Add several commands in a row, using a list of dictionaries
        """
        for command in commands:
            name = command['name']
            func = command['func']
            desc = command.get('desc', '')
            shortdesc = command.get('shortdesc', '')
            completion = command.get('completion')
            usage = command.get('usage', '')
            self.register_command(
                name,
                func,
                desc=desc,
                shortdesc=shortdesc,
                completion=completion,
                usage=usage)

    def register_command(self,
                         name: str,
                         func: Callable[..., Any],
                         *,
                         desc: str = '',
                         shortdesc: str = '',
                         completion: Callable[[windows.Input], Completion | None] | None = None,
                         usage: str = '') -> None:
        """
        Add a command
        """
        if name in self.commands:
            return
        if not desc and shortdesc:
            desc = shortdesc
        self.commands[name] = Command(func, desc, completion, shortdesc, usage)

    def complete_commands(self, the_input: windows.Input) -> bool:
        """
        Does command completion on the specified input for both global and tab-specific
        commands.
        This should be called from the completion method (on tab, for example), passing
        the input where completion is to be made.
        It can completion the command name itself or an argument of the command.
        Returns True if a completion was made, False else.
        """
        txt = the_input.get_text()
        # check if this is a command
        if txt.startswith('/') and not txt.startswith('//'):
            position = the_input.get_argument_position(quoted=False)
            if position == 0:
                words = (
                    ['/%s' % (name) for name in sorted(self.core.commands)]
                    + ['/%s' % (name) for name in sorted(self.commands)]
                )
                the_input.new_completion(words, 0)
                # Do not try to cycle command completion if there was only
                # one possibility. The next tab will complete the argument.
                # Otherwise we would need to add a useless space before being
                # able to complete the arguments.
                hit_copy = set(the_input.hit_list)
                if len(hit_copy) == 1:
                    the_input.do_command(' ')
                    the_input.reset_completion()
                return True
            # check if we are in the middle of the command name
            elif len(txt.split()) > 1 or\
                    (txt.endswith(' ') and not the_input.last_completion):
                command_name = txt.split()[0][1:]
                if command_name in self.commands:
                    command = self.commands[command_name]
                elif command_name in self.core.commands:
                    command = self.core.commands[command_name]
                else:  # Unknown command, cannot complete
                    return False
                if command.comp is None:
                    return False  # There's no completion function
                comp = command.comp(the_input)
                if comp:
                    return comp.run()
        return False

    def execute_command(self, provided_text: str) -> bool:
        """
        Execute the command in the input and return False if
        the input didn't contain a command
        """
        txt = provided_text or self.input.key_enter()
        if txt and txt.startswith('/') and not txt.startswith('//') and\
                not txt.startswith('/me '):
            command = txt.strip().split()[0][1:]
            arg = txt[2 + len(command):]  # jump the '/' and the ' '
            func = None
            if command in self.commands:  # check tab-specific commands
                func = self.commands[command].func
            elif command in self.core.commands:  # check global commands
                func = self.core.commands[command].func
            else:
                low = command.lower()
                if low in self.commands:
                    func = self.commands[low].func
                elif low in self.core.commands:
                    func = self.core.commands[low].func
                else:
                    error_handled = False
                    if self.missing_command_callback is not None:
                        error_handled = self.missing_command_callback(low)
                    if not error_handled:
                        self.information(
                            "Unknown command (%s)" % (command), 'Error')
            if command in ('correct', 'say'):  # hack
                arg = xhtml.convert_simple_to_full_colors(arg)
            else:
                arg = xhtml.clean_text_simple(arg)
            if func:
                if hasattr(self.input, "reset_completion"):
                    self.input.reset_completion()
                result = func(arg)
                if iscoroutine(result):
                    self.core.register_task(result)
            return True
        else:
            return False

    def refresh_tab_win(self) -> None:
        if config.getbool('enable_vertical_tab_list'):
            left_tab_win = self.core.left_tab_win
            if left_tab_win and not self.size.core_degrade_x:
                left_tab_win.refresh()
        elif not self.size.core_degrade_y:
            self.core.tab_win.refresh()

    def refresh_input(self) -> None:
        """Refresh the current input if any"""
        if self.input is not None:
            self.input.refresh()
            self.core.doupdate()

    def refresh(self) -> None:
        """
        Called on each screen refresh (when something has changed)
        """
        pass

    def resize(self) -> None:
        """Called to resize the tab."""
        pass

    def get_name(self) -> str:
        """
        get the name of the tab
        """
        return self.name

    def get_nick(self) -> str:
        """
        Get the nick of the tab (defaults to its name)
        """
        return self.name

    def on_input(self, key: str, raw: bool) -> None:
        """
        raw indicates if the key should activate the associated command or not.
        """
        pass

    def update_commands(self) -> None:
        for c in self.plugin_commands:
            if c not in self.commands:
                self.commands[c] = self.plugin_commands[c]

    def update_keys(self) -> None:
        for k in self.plugin_keys:
            if k not in self.key_func:
                self.key_func[k] = self.plugin_keys[k]

    def on_lose_focus(self) -> None:
        """
        called when this tab loses the focus.
        """
        self.state = 'normal'

    def on_gain_focus(self) -> None:
        """
        called when this tab gains the focus.
        """
        self.state = 'current'

    def on_scroll_down(self) -> bool:
        """
        Defines what happens when we scroll down
        """
        return True

    def on_scroll_up(self) -> bool:
        """
        Defines what happens when we scroll up
        """
        return True

    def on_line_up(self) -> bool:
        """
        Defines what happens when we scroll one line up
        """
        return True

    def on_line_down(self) -> bool:
        """
        Defines what happens when we scroll one line up
        """
        return True

    def on_half_scroll_down(self) -> bool:
        """
        Defines what happens when we scroll half a screen down
        """
        return True

    def on_half_scroll_up(self) -> bool:
        """
        Defines what happens when we scroll half a screen up
        """
        return True

    def on_info_win_size_changed(self) -> None:
        """
        Called when the window with the information is resized
        """
        pass

    def on_close(self) -> None:
        """
        Called when the tab is to be closed
        """
        if self.input:
            self.input.on_delete()
        self.closed = True

    def matching_names(self) -> list[tuple[int, str]]:
        """
        Returns a list of strings that are used to name a tab with the /win
        command.  For example you could switch to a tab that returns
        ['hello', 'coucou'] using /win hel, or /win coucou
        If not implemented in the tab, it just doesn’t match with anything.
        """
        return []

    def __del__(self) -> None:
        log.debug('------ Closing tab %s', self.__class__.__name__)


class GapTab(Tab):
    def __init__(self) -> None:
        return

    def __bool__(self) -> bool:
        return False

    def __len__(self) -> int:
        return 0

    @property
    def name(self) -> str:
        return ''

    def refresh(self) -> None:
        log.debug(
            'WARNING: refresh() called on a gap tab, this should not happen')


class ChatTab(Tab):
    """
    A tab containing a chat of any type.
    Just use this class instead of Tab if the tab needs a recent-words completion
    Also, ^M is already bound to on_enter
    And also, add the /say command
    """
    plugin_commands: dict[str, Command] = {}
    plugin_keys: dict[str, Callable[..., Any]] = {}
    last_sent_message: SMessage | None
    message_type = 'chat'
    timed_event_paused: DelayedEvent | None
    timed_event_not_paused: DelayedEvent | None
    mam_filler: MAMFiller | None
    e2e_encryption: str | None = None
    input: windows.MessageInput

    def __init__(self, core: 'Core', jid: Union[JID, str]) -> None:
        Tab.__init__(self, core)

        self.e2e_encryption = None
        if not isinstance(jid, JID):
            jid = JID(jid)
        assert jid.domain
        self._jid = jid
        #: Is the tab currently requesting MAM data?
        self.query_status = False
        self._name = jid.full
        self.text_win = windows.TextWin()
        self.directed_presence = None
        self._text_buffer = TextBuffer()
        self._text_buffer.add_window(self.text_win)
        self.mam_filler = None
        self.chatstate: str | None = None  # can be "active", "composing", "paused", "gone", "inactive"
        # We keep a reference of the event that will set our chatstate to "paused", so that
        # we can delete it or change it if we need to
        self.timed_event_paused = None
        self.timed_event_not_paused = None
        # Keeps the last sent message to complete it easily in completion_correct
        # and to replace it.
        self.last_sent_message = None
        self.key_func['M-v'] = self.move_separator
        self.key_func['M-h'] = self.scroll_separator
        self.key_func['M-/'] = self.last_words_completion
        self.key_func['^M'] = self.on_enter
        self.key_func['kUP5'] = self.correct_up
        self.register_command(
            'say',
            self.command_say,
            usage='<message>',
            shortdesc='Send the message.')
        self.register_command(
            'scrollback',
            self.command_scrollback,
            usage="end home clear status goto <+|-linecount>|<linenum>|<timestamp>",
            shortdesc='Scrollback to the given line number, message, or clear the buffer.')
        self.commands['sb'] = self.commands['scrollback']
        self.register_command(
            'xhtml',
            self.command_xhtml,
            usage='<custom xhtml>',
            shortdesc='Send custom XHTML.')
        self.register_command(
            'clear', self.command_clear, shortdesc='Clear the current buffer.')
        self.register_command(
            'correct',
            self.command_correct,
            desc='Fix the last message with whatever you want.',
            shortdesc='Correct the last message.',
            completion=self.completion_correct)
        self.chat_state: Optional[str] = None
        self.update_commands()
        self.update_keys()

    def _create_initial_input(self) -> windows.MessageInput:
        return windows.MessageInput()

    @property
    def name(self) -> str:
        if self._name is not None:
            return self._name
        return self._jid.full

    @name.setter
    def name(self, value: Union[JID, str]) -> None:
        if isinstance(value, JID):
            self.jid = value
        elif isinstance(value, str):
            try:
                value = JID(value)
                if value.domain:
                    self._jid = value
            except InvalidJID:
                self._name = str(value)
        else:
            raise TypeError("Name %r must be of type JID or str." % value)

    @property
    def log_name(self) -> str:
        """Name used for the log filename"""
        return self.jid.bare

    @property
    def jid(self) -> JID:
        return JID(self._jid)

    @jid.setter
    def jid(self, value: JID) -> None:
        if not isinstance(value, JID):
            raise TypeError("Jid %r must be of type JID." % value)
        assert value.domain
        self._jid = value

    @property
    def general_jid(self) -> JID:
        raise NotImplementedError

    def log_message(self, message: BaseMessage) -> None:
        """
        Log the messages in the archives.
        """
        if not isinstance(message, LoggableTrait):
            return
        if not logger.log_message(self.log_name, message):
            self.information('Unable to write in the log file', 'Error')

    def add_message(self, message: BaseMessage) -> None:
        self.log_message(message)
        self._text_buffer.add_message(message)

    def modify_message(self,
                       txt: str,
                       old_id: str,
                       new_id: str,
                       time: Optional[datetime],
                       delayed: bool = False,
                       nickname: Optional[str] = None,
                       user: Optional[User] = None,
                       jid: Optional[JID] = None,
                       ) -> bool:
        message = self._text_buffer.modify_message(
            txt, old_id, new_id, user=user, jid=jid, time=time
        )
        if message:
            self.log_message(message)
            self.core.refresh_window()
            return True
        return False

    def last_words_completion(self) -> None:
        """
        Complete the input with words recently said
        """
        # build the list of the recent words
        char_we_dont_want = string.punctuation + ' ’„“”…«»'
        words = []
        for msg in self._text_buffer.messages[:-40:-1]:
            if not msg:
                continue
            txt = xhtml.clean_text(msg.txt)
            for char in char_we_dont_want:
                txt = txt.replace(char, ' ')
            for word in txt.split():
                if len(word) >= 4 and word not in words:
                    words.append(word)
        words.extend([word for word in config.getlist('words') if word])
        self.input.auto_completion(words, ' ', quotify=False)

    def on_enter(self) -> None:
        txt = self.input.key_enter()
        if txt:
            if not self.execute_command(txt):
                if txt.startswith('//'):
                    txt = txt[1:]
                self.core.register_task(
                    self.command_say(xhtml.convert_simple_to_full_colors(txt))
                )
        self.cancel_paused_delay()

    @command_args_parser.raw
    def command_xhtml(self, xhtml: str) -> None:
        """"
        /xhtml <custom xhtml>
        """
        message = self.generate_xhtml_message(xhtml)
        if message:
            message.send()

    def generate_xhtml_message(self, arg: str) -> SMessage | None:
        if not arg:
            return None
        try:
            body = xhtml.clean_text(
                xhtml.xhtml_to_poezio_colors(arg, force=True))
            ET.fromstring(arg)
        except SAXParseException:
            self.information('Could not send custom xhtml', 'Error')
            log.exception('/xhtml: Unable to send custom xhtml')
            return None

        msg: SMessage = self.core.xmpp.make_message(self.get_dest_jid())
        msg['body'] = body
        msg.enable('html')
        msg['html']['body'] = arg
        return msg

    def get_dest_jid(self) -> JID:
        return self.jid

    @refresh_wrapper.always
    @command_args_parser.ignored
    def command_clear(self) -> None:
        """
        /clear
        """
        self._text_buffer.messages = []
        self.text_win.rebuild_everything(self._text_buffer)

    def check_send_chat_state(self) -> bool:
        "If we should send a chat state"
        return True

    def send_chat_state(self, state: str, always_send: bool = False) -> None:
        """
        Send an empty chatstate message
        """
        from poezio.tabs import PrivateTab

        if self.check_send_chat_state():
            if state in ('active', 'inactive',
                         'gone') and self.inactive and not always_send:
                return
            if config.get_by_tabname('send_chat_states', self.general_jid):
                msg: SMessage = self.core.xmpp.make_message(self.get_dest_jid())
                msg['type'] = self.message_type
                msg['chat_state'] = state
                self.chat_state = state
                msg['no-store'] = True
                if isinstance(self, PrivateTab):
                    msg.enable('muc')
                msg.send()

    def send_composing_chat_state(self, empty_after: bool) -> None:
        """
        Send the "active" or "composing" chatstate, depending
        on the the current status of the input
        """
        name = self.general_jid
        if config.get_by_tabname('send_chat_states', name):
            needed = 'inactive' if self.inactive else 'active'
            self.cancel_paused_delay()
            if not empty_after:
                if self.chat_state != "composing":
                    self.send_chat_state("composing")
                self.set_paused_delay(True)
            elif empty_after and self.chat_state != needed:
                self.send_chat_state(needed, True)

    def set_paused_delay(self, _: Any) -> None:
        """
        we create a timed event that will put us to paused
        in a few seconds
        """
        if not config.get_by_tabname('send_chat_states', self.general_jid):
            return
        # First, cancel the delay if it already exists, before rescheduling
        # it at a new date
        self.cancel_paused_delay()
        new_event = timed_events.DelayedEvent(4, self.send_chat_state,
                                              'paused')
        self.core.add_timed_event(new_event)
        self.timed_event_paused = new_event
        new_event = timed_events.DelayedEvent(
            30, self.send_chat_state, 'inactive'
            if self.inactive else 'active')
        self.core.add_timed_event(new_event)
        self.timed_event_not_paused = new_event

    def cancel_paused_delay(self) -> None:
        """
        Remove that event from the list and set it to None.
        Called for example when the input is emptied, or when the message
        is sent
        """
        if self.timed_event_paused is not None:
            self.core.remove_timed_event(self.timed_event_paused)
            self.timed_event_paused = None
            if self.timed_event_not_paused is not None:
                self.core.remove_timed_event(self.timed_event_not_paused)
                self.timed_event_not_paused = None

    def set_last_sent_message(self, msg: SMessage, correct: bool = False) -> None:
        """Ensure last_sent_message is set with the correct attributes"""
        if correct:
            # XXX: Is the copy needed. Is the object passed here reused
            # afterwards? Who knows.
            msg = copy(msg)
            if self.last_sent_message is not None:
                msg['id'] = self.last_sent_message['id']
        self.last_sent_message = msg

    def handle_retract(self, msg: SMessage) -> None:
        """Handle retractions"""
        raise NotImplementedError()

    @command_args_parser.raw
    async def command_correct(self, line: str) -> None:
        """
        /correct <fixed message>
        """
        if not line:
            self.core.command.help('correct')
            return
        if not self.last_sent_message:
            self.information('There is no message to correct.', 'Error')
            return
        await self.command_say(line, correct=True)

    def completion_correct(self, the_input: 'Input') -> Completion | None:
        if self.last_sent_message and the_input.get_argument_position() == 1:
            return Completion(
                the_input.auto_completion, [self.last_sent_message['body']],
                '',
                quotify=False)
        return None

    def correct_up(self) -> None:
        """Go one message up in the correct history"""
        if self.last_sent_message and self.input:
            if isinstance(self.input, windows.HistoryInput):
                self.input.key_up()
            if not self.input.text.startswith('/correct '):
                self.input.text = '/correct ' + self.input.text
            self.input.key_end()

    @property
    def inactive(self) -> bool:
        """Whether we should send inactive or active as a chatstate"""
        return self.core.status.show in ('xa', 'away') or (
            hasattr(self, 'directed_presence')
            and self.directed_presence is not None
            and self.directed_presence
        )

    def move_separator(self) -> None:
        self.text_win.remove_line_separator()
        self.text_win.add_line_separator(self._text_buffer)
        self.text_win.refresh()
        if self.input:
            self.input.refresh()

    def get_conversation_messages(self) -> list[BaseMessage]:
        return self._text_buffer.messages

    def check_scrolled(self) -> None:
        if self.text_win.pos != 0:
            self.state = 'scrolled'

    @command_args_parser.raw
    async def command_say(self, line: str, attention: bool = False, correct: bool = False) -> None:
        pass

    def goto_build_lines(self, new_date: datetime) -> None:
        text_buffer = self._text_buffer
        built_lines = []
        message_count = 0
        timestamp = config.getbool('show_timestamps')
        nick_size = config.getint('max_nick_length')
        for message in text_buffer.messages:
            if not isinstance(message, Message):
                continue
            # Build lines of a message
            txt = message.txt
            nick = truncate_nick(message.nickname, nick_size)
            offset = 0
            theme = get_theme()
            if message.ack:
                if message.ack > 0:
                    offset += poopt.wcswidth(theme.CHAR_ACK_RECEIVED) + 1
                else:
                    offset += poopt.wcswidth(theme.CHAR_NACK) + 1
            if nick:
                offset += poopt.wcswidth(nick) + 2
            if message.revisions > 0:
                offset += ceil(log10(message.revisions + 1))
            if message.me:
                offset += 1
            if timestamp:
                if message.history:
                    offset += 1 + theme.LONG_TIME_FORMAT_LENGTH
            lines = poopt.cut_text(txt, self.text_win.width - offset - 1)
            for line in lines:
                built_lines.append(line)
            # Find the message with timestamp less than or equal to the queried
            # timestamp and goto that location in the tab.
            if message.time <= new_date:
                message_count += 1
                if len(self.text_win.built_lines) - self.text_win.height >= len(built_lines):
                    self.text_win.pos = (
                        len(self.text_win.built_lines)
                        - self.text_win.height - len(built_lines) + 1
                    )
                else:
                    self.text_win.pos = 0
        if message_count == 0:
            self.text_win.scroll_up(len(self.text_win.built_lines))
        self.core.refresh_window()

    @command_args_parser.quoted(0, 2)
    def command_scrollback(self, args: list[str]) -> None:
        """
        /sb clear
        /sb home
        /sb end
        /sb goto <+|-linecount>|<linenum>|<timestamp>
        The format of timestamp must be ‘[dd[.mm]-<days ago>] hh:mi[:ss]’
        """
        if not args:
            args = ['end']
        if len(args) == 1:
            if args[0] == 'end':
                self.text_win.scroll_down(len(self.text_win.built_lines))
                self.core.refresh_window()
                return
            elif args[0] == 'home':
                self.text_win.scroll_up(len(self.text_win.built_lines))
                self.core.refresh_window()
                return
            elif args[0] == 'clear':
                self._text_buffer.messages = []
                self.text_win.rebuild_everything(self._text_buffer)
                self.core.refresh_window()
                return
            elif args[0] == 'status':
                self.information(
                    'Total %s lines in this tab.' % len(self.text_win.built_lines),
                    'Info',
                )
                return
        elif len(args) == 2 and args[0] == 'goto':
            for fmt in ('%d %H:%M', '%d %H:%M:%S', '%d:%m %H:%M',
                        '%d:%m %H:%M:%S', '%H:%M', '%H:%M:%S'):
                try:
                    new_date = datetime.strptime(args[1], fmt)
                    if 'd' in fmt and 'm' in fmt:
                        new_date = new_date.replace(year=datetime.now().year)
                    elif 'd' in fmt:
                        new_date = new_date.replace(
                            year=datetime.now().year,
                            month=datetime.now().month,
                        )
                    else:
                        new_date = new_date.replace(
                            year=datetime.now().year,
                            month=datetime.now().month,
                            day=datetime.now().day,
                        )
                except ValueError:
                    pass
            if args[1].startswith('-'):
                # Check if the user is giving argument of type
                # goto <-linecount> or goto [-<days ago>] hh:mi[:ss]
                if ' ' in args[1]:
                    new_args = args[1].split(' ')
                    new_args[0] = new_args[0].strip('-')
                    new_date = datetime.now()
                    if new_args[0].isdigit():
                        new_date = new_date.replace(day=new_date.day - int(new_args[0]))
                    for fmt in ('%H:%M', '%H:%M:%S'):
                        try:
                            arg_date = datetime.strptime(new_args[1], fmt)
                            new_date = new_date.replace(
                                hour=arg_date.hour,
                                minute=arg_date.minute,
                                second=arg_date.second,
                            )
                        except ValueError:
                            pass
                else:
                    scroll_len = args[1].strip('-')
                    if scroll_len.isdigit():
                        self.text_win.scroll_down(int(scroll_len))
                        self.core.refresh_window()
                        return
            elif args[1].startswith('+'):
                scroll_len = args[1].strip('+')
                if scroll_len.isdigit():
                    self.text_win.scroll_up(int(scroll_len))
                    self.core.refresh_window()
                    return
            # Check for the argument of type goto <linenum>
            elif args[1].isdigit():
                if len(self.text_win.built_lines) - self.text_win.height >= int(args[1]):
                    self.text_win.pos = (
                        len(self.text_win.built_lines)
                        - self.text_win.height - int(args[1])
                    )
                    self.core.refresh_window()
                    return
                else:
                    self.text_win.pos = 0
                    self.core.refresh_window()
                    return
            elif args[1] == '0':
                args = ['home']
            # new_date is the timestamp for which the user has queried.
            self.goto_build_lines(new_date)

    def on_line_up(self) -> bool:
        return self.text_win.scroll_up(1)

    def on_line_down(self) -> bool:
        return self.text_win.scroll_down(1)

    def on_scroll_up(self) -> bool:
        if not self.query_status:
            from poezio.log_loader import LogLoader
            self.core.register_task(
                LogLoader(logger, self, config.getbool('use_log')).scroll_requested()
            )
        return self.text_win.scroll_up(self.text_win.height - 1)

    def on_scroll_down(self) -> bool:
        return self.text_win.scroll_down(self.text_win.height - 1)

    def on_half_scroll_up(self) -> bool:
        return self.text_win.scroll_up((self.text_win.height - 1) // 2)

    def on_half_scroll_down(self) -> bool:
        return self.text_win.scroll_down((self.text_win.height - 1) // 2)

    @refresh_wrapper.always
    def scroll_separator(self) -> None:
        self.text_win.scroll_to_separator()


class OneToOneTab(ChatTab):
    def __init__(self, core: 'Core', jid: JID, initial: SMessage | None = None) -> None:
        ChatTab.__init__(self, core, jid)

        self.__status = Status("", "")
        self.last_remote_message = datetime.now()
        self._initial_log = asyncio.Event()

        # Set to true once the first disco is done
        self.__initial_disco = False
        self.register_command(
            'unquery', self.command_unquery, shortdesc='Close the tab.')
        self.register_command(
            'close', self.command_unquery, shortdesc='Close the tab.')
        self.register_command(
            'attention',
            self.command_attention,
            usage='[message]',
            shortdesc='Request the attention.',
            desc='Attention: Request the attention of the contact.  Can also '
            'send a message along with the attention.')
        self.core.register_task(self.init_logs(initial=initial))

    async def init_logs(self, initial: Optional[SMessage] = None) -> None:
        """
        Load logs on tab creation. No MAM code because this is shared with
        the private conversations tab.
        """
        use_log = config.get_by_tabname('use_log', self.jid)
        task = self.core.register_task(
            LogLoader(logger, self, use_log, self._initial_log).tab_open()
        )
        await asyncio.wait([task], timeout=5)
        if initial:
            await self.handle_message(initial)

    async def handle_message(self, msg: SMessage, display: bool = True) -> None:
        pass

    def refresh_info_header(self) -> None:
        pass

    @refresh_wrapper.conditional
    def handle_reactions(self, message: SMessage) -> bool:
        reactions = message['reactions']['values']
        react_id = message['reactions']['id']
        return self._text_buffer.react(
            by=message['from'].bare,
            to=react_id,
            reactions=reactions,
        )

    def remote_user_color(self) -> str:
        return dump_tuple(get_theme().COLOR_REMOTE_USER)

    def update_status(self, status: Status) -> None:
        old_status = self.__status
        if not (old_status.show != status.show
                or old_status.message != status.message):
            return
        self.__status = status
        hide_status_change = config.get_by_tabname('hide_status_change',
                                                   JID(self.jid.bare))
        now = datetime.now()
        dff = now - self.last_remote_message
        if hide_status_change > -1 and dff.total_seconds() > hide_status_change:
            return

        info_c = dump_tuple(get_theme().COLOR_INFORMATION_TEXT)
        nick = self.get_nick()
        remote = self.remote_user_color()
        msg = '\x19%(color)s}%(nick)s\x19%(info)s} changed: '
        msg %= {'color': remote, 'nick': nick, 'info': info_c}
        if status.message != old_status.message and status.message:
            msg += 'status: %s, ' % status.message
        if status.show in SHOW_NAME:
            msg += 'show: %s, ' % SHOW_NAME[status.show]
        self.add_message(
            PersistentInfoMessage(txt=msg[:-2])
        )

    def handle_retract(self, msg: SMessage) -> None:
        """Handler called upon receiving a message retraction"""
        to_retract = msg['retract']['id']
        jid = msg['from']
        self.handle_retract_values(to_retract, jid)

    def handle_retract_values(self, to_retract: str, jid: JID) -> None:
        """Sub-handler used to handle message retractions without a stanza
        object.
        :param to_retract: id of the message to retract.
        :param jid: JID the retraction is coming from.
        """
        try:
            self._text_buffer.delete_message(
                orig_id=to_retract,
                jid=jid,
                muc=False,
                tombstone='This message has been retracted.',
            )
        except DeletionError as exc:
            err = exc.args[0]
            self.information(err, 'Warning')
            log.exception('Failed to retract a message:')
        self.core.refresh_window()

    def ack_message(self, msg_id: str, msg_jid: JID) -> None:
        """
        Ack a message
        """
        new_msg = self._text_buffer.ack_message(msg_id, msg_jid)
        if new_msg:
            self.core.refresh_window()

    def nack_message(self, error: str, msg_id: str, msg_jid: JID) -> bool:
        """
        Non-ack a message (e.g. timeout)
        """
        new_msg = self._text_buffer.nack_message(error, msg_id, msg_jid)
        if new_msg:
            self.core.refresh_window()
            return True
        return False

    @command_args_parser.raw
    def command_xhtml(self, xhtml_str: str) -> None:
        message = self.generate_xhtml_message(xhtml_str)
        if message:
            message['type'] = 'chat'
            # This is our own hack, which makes sense to type ignore
            message._add_receipt = True  # type: ignore
            message['chat_sate'] = 'active'
            message.send()
            body = xhtml.xhtml_to_poezio_colors(xhtml_str, force=True)
            self._text_buffer.add_message(
                Message(
                    body,
                    nickname=self.core.own_nick,
                    nick_color=get_theme().COLOR_OWN_NICK,
                    identifier=message['id'],
                    jid=self.core.xmpp.boundjid,
                )
            )
            self.refresh()

    @command_args_parser.raw
    async def command_attention(self, message: str) -> None:
        """/attention [message]"""
        if message != '':
            await self.command_say(message, attention=True)
        else:
            msg = self.core.xmpp.make_message(self.get_dest_jid())
            msg['type'] = 'chat'
            msg['attention'] = True
            msg.send()

    @command_args_parser.raw
    async def command_say(self, line: str, attention: bool = False, correct: bool = False) -> None:
        pass

    @command_args_parser.ignored
    def command_unquery(self) -> None:
        """
        /unquery
        """
        self.send_chat_state('gone', always_send=True)
        self.core.close_tab(self)

    def missing_command_callback(self, command_name: str) -> bool:
        if command_name not in ('correct', 'attention'):
            return False

        feature = ''
        if command_name == 'correct':
            feature = 'message correction'
        elif command_name == 'attention':
            feature = 'attention requests'
        if feature:
            msg = ('%s does not support %s, therefore the /%s '
                   'command is currently disabled in this tab.')
            msg = msg % (self.name, feature, command_name)
            self.information(msg, 'Info')
        return True


class DefaultHelpTab(Tab):
    """
    A tab showing an help message on the bottom, but pressing / allows to
    enter a command.
    """
    default_help_message: windows.HelpText
    input: windows.HelpText | windows.CommandInput

    def _create_initial_input(self) -> windows.HelpText:
        return self.default_help_message

    def execute_slash_command(self, txt: str) -> bool:
        if txt.startswith('/') and self.input is not None:
            self.input.key_enter()
            self.execute_command(txt)
        return self.reset_help_message()

    @refresh_wrapper.always
    def reset_help_message(self, _: Any = None) -> bool:
        if self.closed:
            return True
        self.input = self.default_help_message
        if self.core.tabs.current_tab is self:
            curses.curs_set(0)
            self.input.resize(1, self.width, self.height - 1, 0)
        return True

    def on_slash(self) -> None:
        """
        '/' is pressed, we enter "input mode"
        """
        curses.curs_set(1)
        self.input = windows.CommandInput("", self.reset_help_message,
                                          self.execute_slash_command)
        self.input.resize(1, self.width, self.height - 1, 0)
        self.input.do_command("/")  # we add the slash

    def completion(self) -> None:
        """Complete the input if it is not the help text."""
        if isinstance(self.input, windows.Input):
            self.complete_commands(self.input)

    def on_input(self, key: str, raw: bool) -> None:
        res = self.input.do_command(key, raw=raw)
        if res:
            return
        if not raw and key in self.key_func:
            self.key_func[key]()


class NoInputTab(Tab):
    """A tab that does not allow any input in the bottom text bar whatsoever."""
    def _create_initial_input(self) -> windows.NoInput:
        return windows.NoInput()



"""
Defines a window which contains either an image or a border.
"""

from __future__ import annotations

import curses
from io import BytesIO
from typing import Union

try:
    from PIL import Image, ImageFile
    HAS_PIL = True
except ImportError:
    HAS_PIL = False

try:
    import gi
    gi.require_version('Rsvg', '2.0')
    from gi.repository import Rsvg
    import cairo
    HAS_RSVG = True
except (ImportError, ValueError, AttributeError):
    HAS_RSVG = False

from poezio.windows.base_wins import Win
from poezio.theming import get_theme, to_curses_attr
from poezio.xhtml import _parse_css_color
from poezio.config import config

from typing import Tuple, Optional, Callable


MAX_SIZE = 16


def render_svg(svg: bytes) -> Optional[Image.Image]:
    if not HAS_RSVG:
        return None
    try:
        handle = Rsvg.Handle.new_from_data(svg)
        if handle is None:
            return None
        dimensions = handle.get_dimensions()
        biggest_dimension = max(dimensions.width, dimensions.height)
        scale = MAX_SIZE / biggest_dimension
        translate_x = (biggest_dimension - dimensions.width) / 2
        translate_y = (biggest_dimension - dimensions.height) / 2

        surface = cairo.ImageSurface(cairo.Format.ARGB32, MAX_SIZE, MAX_SIZE)
        context = cairo.Context(surface)
        context.scale(scale, scale)
        context.translate(translate_x, translate_y)
        handle.render_cairo(context)
        data = surface.get_data()
        image = Image.frombytes('RGBA', (MAX_SIZE, MAX_SIZE), data.tobytes())
        # This is required because Cairo uses a BGRA (in host endianness)
        # format, and PIL an ABGR (in byte order) format.  Yes, this is
        # confusing.
        b, g, r, a = image.split()
        return Image.merge('RGB', (r, g, b))  # noqa: TRY300
    except Exception:
        return None


class ImageWin(Win):
    """
    A window which contains either an image or a border.
    """

    __slots__ = ('_image', '_display_avatar')

    def __init__(self) -> None:
        self._image: Optional[Image.Image] = None
        Win.__init__(self)
        if config.getbool('image_use_half_blocks'):
            self._display_avatar: Callable[[int, int], None] = self._display_avatar_half_blocks
        else:
            self._display_avatar = self._display_avatar_full_blocks

    def resize(self, height: int, width: int, y: int, x: int) -> None:
        self._resize(height, width, y, x)
        if self._image is None:
            return
        self._display_avatar(width, height)

    def refresh(self, data: Optional[bytes]) -> None:
        self._win.erase()
        if data is not None and HAS_PIL:
            image_file = BytesIO(data)
            image: Union[Image.Image, ImageFile.ImageFile, None]
            try:
                try:
                    image = Image.open(image_file)
                except OSError:
                    # TODO: Make the caller pass the MIME type, so we don’t
                    # have to try all renderers like that.
                    image = render_svg(data)
                    if image is None:
                        raise
            except OSError:
                self._display_border()
            else:
                self._image = image.convert('RGB')
                self._display_avatar(self.width, self.height)
        else:
            self._display_border()
        self._refresh()

    def _display_border(self) -> None:
        self._image = None
        attribute = to_curses_attr(get_theme().COLOR_VERTICAL_SEPARATOR)
        self._win.attron(attribute)
        self._win.border(curses.ACS_VLINE, curses.ACS_VLINE, curses.ACS_HLINE,
                         curses.ACS_HLINE, curses.ACS_ULCORNER,
                         curses.ACS_URCORNER, curses.ACS_LLCORNER,
                         curses.ACS_LRCORNER)
        self._win.attroff(attribute)

    @staticmethod
    def _compute_size(image_size: Tuple[int, int], width: int, height: int) -> Tuple[int, int]:
        height *= 2
        src_width, src_height = image_size
        ratio = src_width / src_height
        new_width = height * ratio
        new_height = width / ratio
        if new_width > width:
            height = int(new_height)
        elif new_height > height:
            width = int(new_width)
        return width, height

    def _display_avatar_half_blocks(self, width: int, height: int) -> None:
        if self._image is None:
            return
        original_height = height
        original_width = width
        size = self._compute_size(self._image.size, width, height)
        image2 = self._image.resize(size, resample=Image.BILINEAR)  # type: ignore # pylint: disable=no-member
        data = image2.tobytes()
        width, height = size
        start_y = (original_height - height // 2) // 2
        start_x = (original_width - width) // 2
        for y in range(height // 2):
            two_lines = data[(2 * y) * width * 3:(2 * y + 2) * width * 3]
            line1 = two_lines[:width * 3]
            line2 = two_lines[width * 3:]
            self.move(start_y + y, start_x)
            for x in range(0, width * 3, 3):
                r, g, b = line1[x:x + 3]
                top_color = _parse_css_color('#%02x%02x%02x' % (r, g, b))
                r, g, b = line2[x:x + 3]
                bot_color = _parse_css_color('#%02x%02x%02x' % (r, g, b))
                self.addstr('▄', to_curses_attr((bot_color, top_color)))

    def _display_avatar_full_blocks(self, width: int, height: int) -> None:
        if self._image is None:
            return
        original_height = height
        original_width = width
        width, height = self._compute_size(self._image.size, width, height)
        height //= 2
        size = width, height
        image2 = self._image.resize(size, resample=Image.BILINEAR)  # type: ignore # pylint: disable=no-member
        data = image2.tobytes()
        start_y = (original_height - height) // 2
        start_x = (original_width - width) // 2
        for y in range(height):
            line = data[y * width * 3:(y + 1) * width * 3]
            self.move(start_y + y, start_x)
            for x in range(0, width * 3, 3):
                r, g, b = line[x:x + 3]
                color = _parse_css_color('#%02x%02x%02x' % (r, g, b))
                self.addstr('█', to_curses_attr((color, -1)))



"""
Dice plugin: roll some dice

Usage of this plugin is not recommended.

Commands
--------

.. glossary::

    /roll [number of dice] [duration of the roll]
        Roll one or several unicode dice

Configuration
-------------

.. glossary::
    :sorted:

    refresh [dice plugin]
        **Default:** ``0.5``

        Interval in seconds between each correction (the closest to 0 is the fastest)

    default_duration
        **Default:** ``5``

        Total duration of the animation.
"""

import random

from poezio import tabs
from poezio.decorators import command_args_parser
from poezio.plugin import BasePlugin
from slixmpp import JID

DICE = '\u2680\u2681\u2682\u2683\u2684\u2685'


class DiceRoll:
    __slots__ = [
        'duration', 'total_duration', 'dice_number', 'msgtype', 'jid',
        'msgid', 'increments'
    ]

    def __init__(self, total_duration: int, dice_number: int, msgtype: str,
                 jid: JID, msgid: str, increments: int) -> None:
        self.duration = 0
        self.total_duration = total_duration
        self.dice_number = dice_number
        self.msgtype = msgtype
        self.jid = jid
        self.msgid = msgid
        self.increments = increments

    def reroll(self) -> None:
        self.duration += self.increments

    def is_finished(self) -> bool:
        return self.duration >= self.total_duration


def roll_dice(num_dice: int) -> str:
    return ''.join(random.choice(DICE) for _ in range(num_dice))


class Plugin(BasePlugin):
    default_config = {"dice": {"refresh": 0.75, "default_duration": 7.5}}

    def init(self) -> None:
        for tab_t in [tabs.MucTab, tabs.DynamicConversationTab, tabs.StaticConversationTab, tabs.PrivateTab]:
            self.api.add_tab_command(
                tab_t,
                'roll',
                self.command_dice,
                help='Roll a die',
                usage='[number] [duration]')

    @command_args_parser.quoted(0, 2, ['', ''], True)
    def command_dice(self, args: list[str]) -> None:
        tab = self.api.current_tab()
        if not isinstance(tab, tabs.ChatTab):
            return
        duration = self.config.get('default_duration')
        num_dice = 1
        try:
            if args[0]:
                num_dice = int(args[0])
                if args[1]:
                    duration = float(args[1])
        except ValueError:
            self.core.command.help("roll")
            return
        else:
            if num_dice <= 0 or duration < 0:
                self.core.command.help("roll")
                return

        msgtype = 'groupchat' if isinstance(tab, tabs.MucTab) else 'chat'

        message = self.core.xmpp.make_message(tab.jid)
        message['type'] = msgtype
        message['body'] = roll_dice(num_dice)
        message.send()

        increment = self.config.get('refresh')
        msgid = message['id']

        roll = DiceRoll(duration, num_dice, msgtype, tab.jid, msgid, increment)
        event = self.api.create_delayed_event(increment, self.delayed_event,
                                              roll)
        self.api.add_timed_event(event)

    def delayed_event(self, roll: DiceRoll) -> None:
        if roll.is_finished():
            return
        roll.reroll()
        message = self.core.xmpp.make_message(roll.jid)
        message["type"] = roll.msgtype
        message["body"] = roll_dice(roll.dice_number)
        message["replace"]["id"] = roll.msgid
        message.send()
        event = self.api.create_delayed_event(roll.increments,
                                              self.delayed_event, roll)
        self.api.add_timed_event(event)



# This file was taken from https://github.com/hsluv/hsluv-python
#
# Copyright (c) 2015 Alexei Boronine
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
""" This module is generated by transpiling Haxe into Python and cleaning
the resulting code by hand, e.g. removing unused Haxe classes. To try it
yourself, clone https://github.com/hsluv/hsluv and run:

    haxe -cp haxe/src hsluv.Hsluv -python hsluv.py
"""

import math


__version__ = '0.0.2'

m = [[3.240969941904521, -1.537383177570093, -0.498610760293],
     [-0.96924363628087, 1.87596750150772, 0.041555057407175],
     [0.055630079696993, -0.20397695888897, 1.056971514242878]]
minv = [[0.41239079926595, 0.35758433938387, 0.18048078840183],
        [0.21263900587151, 0.71516867876775, 0.072192315360733],
        [0.019330818715591, 0.11919477979462, 0.95053215224966]]
refY = 1.0
refU = 0.19783000664283
refV = 0.46831999493879
kappa = 903.2962962
epsilon = 0.0088564516
hex_chars = "0123456789abcdef"


def _distance_line_from_origin(line):
    v = math.pow(line['slope'], 2) + 1
    return math.fabs(line['intercept']) / math.sqrt(v)


def _length_of_ray_until_intersect(theta, line):
    return line['intercept'] / (math.sin(theta) - line['slope'] * math.cos(theta))


def _get_bounds(l):
    result = []
    sub1 = math.pow(l + 16, 3) / 1560896
    if sub1 > epsilon:
        sub2 = sub1
    else:
        sub2 = l / kappa
    _g = 0
    while _g < 3:
        c = _g
        _g = _g + 1
        m1 = m[c][0]
        m2 = m[c][1]
        m3 = m[c][2]
        _g1 = 0
        while _g1 < 2:
            t = _g1
            _g1 = _g1 + 1
            top1 = (284517 * m1 - 94839 * m3) * sub2
            top2 = (838422 * m3 + 769860 * m2 + 731718 * m1) * l * sub2 - (769860 * t) * l
            bottom = (632260 * m3 - 126452 * m2) * sub2 + 126452 * t
            result.append({'slope': top1 / bottom, 'intercept': top2 / bottom})
    return result


def _max_safe_chroma_for_l(l):
    bounds = _get_bounds(l)
    _hx_min = 1.7976931348623157e+308
    _g = 0
    while _g < 2:
        i = _g
        _g = _g + 1
        length = _distance_line_from_origin(bounds[i])
        if math.isnan(_hx_min):
            _hx_min = _hx_min
        elif math.isnan(length):
            _hx_min = length
        else:
            _hx_min = min(_hx_min, length)
    return _hx_min


def _max_chroma_for_lh(l, h):
    hrad = h / 360 * math.pi * 2
    bounds = _get_bounds(l)
    _hx_min = 1.7976931348623157e+308
    _g = 0
    while _g < len(bounds):
        bound = bounds[_g]
        _g = (_g + 1)
        length = _length_of_ray_until_intersect(hrad, bound)
        if length >= 0:
            if math.isnan(_hx_min):
                _hx_min = _hx_min
            elif math.isnan(length):
                _hx_min = length
            else:
                _hx_min = min(_hx_min, length)
    return _hx_min


def _dot_product(a, b):
    sum = 0
    _g1 = 0
    _g = len(a)
    while _g1 < _g:
        i = _g1
        _g1 = _g1 + 1
        sum += a[i] * b[i]
    return sum


def _from_linear(c):
    if c <= 0.0031308:
        return 12.92 * c
    else:
        return 1.055 * math.pow(c, 0.416666666666666685) - 0.055


def _to_linear(c):
    if c > 0.04045:
        return math.pow((c + 0.055) / 1.055, 2.4)
    else:
        return c / 12.92


def xyz_to_rgb(_hx_tuple):
    return [
        _from_linear(_dot_product(m[0], _hx_tuple)),
        _from_linear(_dot_product(m[1], _hx_tuple)),
        _from_linear(_dot_product(m[2], _hx_tuple))]


def rgb_to_xyz(_hx_tuple):
    rgbl = [_to_linear(_hx_tuple[0]),
            _to_linear(_hx_tuple[1]),
            _to_linear(_hx_tuple[2])]
    return [_dot_product(minv[0], rgbl),
            _dot_product(minv[1], rgbl),
            _dot_product(minv[2], rgbl)]


def _y_to_l(y):
    if y <= epsilon:
        return y / refY * kappa
    else:
        return 116 * math.pow(y / refY, 0.333333333333333315) - 16


def _l_to_y(l):
    if l <= 8:
        return refY * l / kappa
    else:
        return refY * math.pow((l + 16) / 116, 3)


def xyz_to_luv(_hx_tuple):
    x = float(_hx_tuple[0])
    y = float(_hx_tuple[1])
    z = float(_hx_tuple[2])
    divider = x + 15 * y + 3 * z
    var_u = 4 * x
    var_v = 9 * y
    if divider != 0:
        var_u = var_u / divider
        var_v = var_v / divider
    else:
        var_u = float("nan")
        var_v = float("nan")
    l = _y_to_l(y)
    if l == 0:
        return [0, 0, 0]
    u = 13 * l * (var_u - refU)
    v = 13 * l * (var_v - refV)
    return [l, u, v]


def luv_to_xyz(_hx_tuple):
    l = float(_hx_tuple[0])
    u = float(_hx_tuple[1])
    v = float(_hx_tuple[2])
    if l == 0:
        return [0, 0, 0]
    var_u = u / (13 * l) + refU
    var_v = v / (13 * l) + refV
    y = _l_to_y(l)
    x = 0 - ((9 * y * var_u) / (((var_u - 4) * var_v) - var_u * var_v))
    z = (((9 * y) - (15 * var_v * y)) - (var_v * x)) / (3 * var_v)
    return [x, y, z]


def luv_to_lch(_hx_tuple):
    l = float(_hx_tuple[0])
    u = float(_hx_tuple[1])
    v = float(_hx_tuple[2])
    _v = (u * u) + (v * v)
    if _v < 0:
        c = float("nan")
    else:
        c = math.sqrt(_v)
    if c < 0.00000001:
        h = 0
    else:
        hrad = math.atan2(v, u)
        h = hrad * 180.0 / 3.1415926535897932
        if h < 0:
            h = 360 + h
    return [l, c, h]


def lch_to_luv(_hx_tuple):
    l = float(_hx_tuple[0])
    c = float(_hx_tuple[1])
    h = float(_hx_tuple[2])
    hrad = h / 360.0 * 2 * math.pi
    u = math.cos(hrad) * c
    v = math.sin(hrad) * c
    return [l, u, v]


def hsluv_to_lch(_hx_tuple):
    h = float(_hx_tuple[0])
    s = float(_hx_tuple[1])
    l = float(_hx_tuple[2])
    if l > 99.9999999:
        return [100, 0, h]
    if l < 0.00000001:
        return [0, 0, h]
    _hx_max = _max_chroma_for_lh(l, h)
    c = _hx_max / 100 * s
    return [l, c, h]


def lch_to_hsluv(_hx_tuple):
    l = float(_hx_tuple[0])
    c = float(_hx_tuple[1])
    h = float(_hx_tuple[2])
    if l > 99.9999999:
        return [h, 0, 100]
    if l < 0.00000001:
        return [h, 0, 0]
    _hx_max = _max_chroma_for_lh(l, h)
    s = c / _hx_max * 100
    return [h, s, l]


def hpluv_to_lch(_hx_tuple):
    h = float(_hx_tuple[0])
    s = float(_hx_tuple[1])
    l = float(_hx_tuple[2])
    if l > 99.9999999:
        return [100, 0, h]
    if l < 0.00000001:
        return [0, 0, h]
    _hx_max = _max_safe_chroma_for_l(l)
    c = _hx_max / 100 * s
    return [l, c, h]


def lch_to_hpluv(_hx_tuple):
    l = float(_hx_tuple[0])
    c = float(_hx_tuple[1])
    h = float(_hx_tuple[2])
    if l > 99.9999999:
        return [h, 0, 100]
    if l < 0.00000001:
        return [h, 0, 0]
    _hx_max = _max_safe_chroma_for_l(l)
    s = c / _hx_max * 100
    return [h, s, l]


def rgb_to_hex(_hx_tuple):
    h = "#"
    _g = 0
    while _g < 3:
        i = _g
        _g = _g + 1
        chan = float(_hx_tuple[i])
        c = math.floor(chan * 255 + 0.5)
        digit2 = int(c % 16)
        digit1 = int((c - digit2) / 16)

        h += hex_chars[digit1] + hex_chars[digit2]
    return h


def hex_to_rgb(hex):
    hex = hex.lower()
    ret = []
    _g = 0
    while _g < 3:
        i = _g
        _g = _g + 1
        index = i * 2 + 1
        _hx_str = hex[index]
        digit1 = hex_chars.find(_hx_str)
        index1 = i * 2 + 2
        str1 = hex[index1]
        digit2 = hex_chars.find(str1)
        n = digit1 * 16 + digit2
        ret.append(n / 255.0)
    return ret


def lch_to_rgb(_hx_tuple):
    return xyz_to_rgb(luv_to_xyz(lch_to_luv(_hx_tuple)))


def rgb_to_lch(_hx_tuple):
    return luv_to_lch(xyz_to_luv(rgb_to_xyz(_hx_tuple)))


def hsluv_to_rgb(_hx_tuple):
    return lch_to_rgb(hsluv_to_lch(_hx_tuple))


def rgb_to_hsluv(_hx_tuple):
    return lch_to_hsluv(rgb_to_lch(_hx_tuple))


def hpluv_to_rgb(_hx_tuple):
    return lch_to_rgb(hpluv_to_lch(_hx_tuple))


def rgb_to_hpluv(_hx_tuple):
    return lch_to_hpluv(rgb_to_lch(_hx_tuple))


def hsluv_to_hex(_hx_tuple):
    return rgb_to_hex(hsluv_to_rgb(_hx_tuple))


def hpluv_to_hex(_hx_tuple):
    return rgb_to_hex(hpluv_to_rgb(_hx_tuple))


def hex_to_hsluv(s):
    return rgb_to_hsluv(hex_to_rgb(s))


def hex_to_hpluv(s):
    return rgb_to_hpluv(hex_to_rgb(s))



from typing import Callable, TypedDict, Any

from poezio.core.commands import CommandCore
from poezio.core.structs import GenericCompletion
from poezio.core.completions import CompletionCore
from poezio.plugin_manager import PluginManager


CommandDict = TypedDict(
    "CommandDict",
    {
        "name": str,
        "func": Callable[..., Any],
        "shortdesc": str,
        "desc": str,
        "usage": str,
        "completion": GenericCompletion,
    },
    total=False,
)


def get_commands(commands: CommandCore, completions: CompletionCore,
                 plugin_manager: PluginManager) -> list[CommandDict]:
    """
    Get the set of default poezio commands.
    """
    return [
        {
            "name": "help",
            "func": commands.help,
            "usage": "[command]",
            "shortdesc": "\\_o< KOIN KOIN KOIN",
            "completion": completions.help,
        },
        {
            "name": "join",
            "func": commands.join,
            "usage": "[room_name][@server][/nick] [password]",
            "desc": (
                "Join the specified room. You can specify a nickname "
                "after a slash (/). If no nickname is specified, you will"
                " use the default_nick in the configuration file. You can"
                " omit the room name: you will then join the room you're"
                " looking at (useful if you were kicked). You can also "
                "provide a room_name without specifying a server, the "
                "server of the room you're currently in will be used. You"
                " can also provide a password to join the room.\nExamples"
                ":\n/join room@server.tld\n/join room@server.tld/John\n"
                "/join room2\n/join /me_again\n/join\n/join room@server"
                ".tld/my_nick password\n/join / password"
            ),
            "shortdesc": "Join a room",
            "completion": completions.join,
        },
        {
            "name": "exit",
            "func": commands.quit,
            "desc": "Just disconnect from the server and exit poezio.",
            "shortdesc": "Exit poezio.",
        },
        {
            "name": "quit",
            "func": commands.quit,
            "desc": "Just disconnect from the server and exit poezio.",
            "shortdesc": "Exit poezio.",
        },
        {
            "name": "next",
            "func": commands.rotate_rooms_right,
            "shortdesc": "Go to the next room.",
        },
        {
            "name": "prev",
            "func": commands.rotate_rooms_left,
            "shortdesc": "Go to the previous room.",
        },
        {
            "name": "win",
            "func": commands.win,
            "usage": "<number or name>",
            "shortdesc": "Go to the specified room",
            "completion": completions.win,
        },
        {
            "name": "w",
            "func": commands.win,
            "usage": "<number or name>",
            "shortdesc": "Go to the specified room",
            "completion": completions.win,
        },
        {
            "name": "wup",
            "func": commands.wup,
            "usage": "<prefix>",
            "shortdesc": "Go to the tab whose name uniquely starts with prefix",
            "completion": completions.win,
        },
        {
            "name": "move_tab",
            "func": commands.move_tab,
            "usage": "<source> <destination>",
            "desc": (
                "Insert the <source> tab at the position of "
                "<destination>. This will make the following tabs shift in"
                " some cases (refer to the documentation). A tab can be "
                "designated by its number or by the beginning of its "
                'address. You can use "." as a shortcut for the current '
                "tab."
            ),
            "shortdesc": "Move a tab.",
            "completion": completions.move_tab,
        },
        {
            "name": "destroy_room",
            "func": commands.destroy_room,
            "usage": "[room JID]",
            "desc": (
                "Try to destroy the room [room JID], or the current"
                " tab if it is a multi-user chat and [room JID] is "
                "not given."
            ),
            "shortdesc": "Destroy a room.",
            "completion": None,
        },
        {
            "name": "report",
            "func": commands.report,
            "usage": "[JID]",
            "desc": (
                "Reports the specified account to the server."
            ),
            "shortdesc": "Report an account.",
            "completion": completions.report,
        },
        {
            "name": "status",
            "func": commands.status,
            "usage": "<availability> [status message]",
            "desc": (
                "Sets your availability and (optionally) your status "
                'message. The <availability> argument is one of "available'
                ', chat, away, afk, dnd, busy, xa" and the optional '
                "[status message] argument will be your status message."
            ),
            "shortdesc": "Change your availability.",
            "completion": completions.status,
        },
        {
            "name": "show",
            "func": commands.status,
            "usage": "<availability> [status message]",
            "desc": (
                "Sets your availability and (optionally) your status "
                'message. The <availability> argument is one of "available'
                ', chat, away, afk, dnd, busy, xa" and the optional '
                "[status message] argument will be your status message."
            ),
            "shortdesc": "Change your availability.",
            "completion": completions.status,
        },
        {
            "name": "bookmark_local",
            "func": commands.bookmark_local,
            "usage": "[roomname][/nick] [password]",
            "desc": (
                "Bookmark Local: Bookmark locally the specified room "
                "(you will then auto-join it on each poezio start). This"
                " commands uses almost the same syntax as /join. Type "
                "/help join for syntax examples. Note that when typing "
                '"/bookmark" on its own, the room will be bookmarked '
                "with the nickname you're currently using in this room "
                "(instead of default_nick)"
            ),
            "shortdesc": "Bookmark a room locally.",
            "completion": completions.bookmark_local,
        },
        {
            "name": "bookmark",
            "func": commands.bookmark,
            "usage": "[roomname][/nick] [autojoin] [password]",
            "desc": (
                "Bookmark: Bookmark online the specified room (you "
                "will then auto-join it on each poezio start if autojoin"
                " is specified and is 'true'). This commands uses almost"
                " the same syntax as /join. Type /help join for syntax "
                'examples. Note that when typing "/bookmark" alone, the'
                " room will be bookmarked with the nickname you're "
                "currently using in this room (instead of default_nick)."
            ),
            "shortdesc": "Bookmark a room online.",
            "completion": completions.bookmark,
        },
        {
            "name": "accept",
            "func": commands.accept,
            "usage": "[jid]",
            "desc": (
                "Allow the provided JID (or the selected contact "
                "in your roster), to see your presence."
            ),
            "shortdesc": "Allow a user your presence.",
            "completion": completions.roster_barejids,
        },
        {
            "name": "add",
            "func": commands.add,
            "usage": "<jid>",
            "desc": (
                "Add the specified JID to your roster, ask them to"
                " allow you to see his presence, and allow them to"
                " see your presence."
            ),
            "shortdesc": "Add a user to your roster.",
        },
        {
            "name": "deny",
            "func": commands.deny,
            "usage": "[jid]",
            "desc": (
                "Deny your presence to the provided JID (or the "
                "selected contact in your roster), who is asking"
                "you to be in their roster."
            ),
            "shortdesc": "Deny a user your presence.",
            "completion": completions.roster_barejids,
        },
        {
            "name": "remove",
            "func": commands.remove,
            "usage": "[jid]",
            "desc": (
                "Remove the specified JID from your roster. This "
                "will unsubscribe you from its presence, cancel "
                "its subscription to yours, and remove the item "
                "from your roster."
            ),
            "shortdesc": "Remove a user from your roster.",
            "completion": completions.remove,
        },
        {
            "name": "reconnect",
            "func": commands.command_reconnect,
            "usage": "[reconnect]",
            "desc": (
                "Disconnect from the remote server if you are "
                "currently connected and then connect to it again."
            ),
            "shortdesc": "Disconnect and reconnect to the server.",
        },
        {
            "name": "set",
            "func": commands.set_cmd,
            "usage": "[plugin|][section] <option> [value]",
            "desc": (
                "Set the value of an option in your configuration file."
                " You can, for example, change your default nickname by "
                "doing `/set default_nick toto` or your resource with `/set"
                " resource blabla`. You can also set options in specific "
                "sections with `/set bindings M-i ^i` or in specific plugin"
                " with `/set mpd_client| host 127.0.0.1`. `toggle` can be "
                "used as a special value to toggle a boolean option."
            ),
            "shortdesc": "Set the value of an option",
            "completion": completions.set,
        },
        {
            "name": "set_default",
            "func": commands.set_default,
            "usage": "[section] <option>",
            "desc": (
                "Set the default value of an option. For example, "
                "`/set_default resource` will reset the resource "
                "option. You can also reset options in specific "
                "sections by doing `/set_default section option`."
            ),
            "shortdesc": "Set the default value of an option",
            "completion": completions.set_default,
        },
        {
            "name": "toggle",
            "func": commands.toggle,
            "usage": "<option>",
            "desc": "Shortcut for /set <option> toggle",
            "shortdesc": "Toggle an option",
            "completion": completions.toggle,
        },
        {
            "name": "theme",
            "func": commands.theme,
            "usage": "[theme name]",
            "desc": (
                "Reload the theme defined in the config file. If theme"
                "_name is provided, set that theme before reloading it."
            ),
            "shortdesc": "Load a theme",
            "completion": completions.theme,
        },
        {
            "name": "list",
            "func": commands.list_cmd,
            "usage": "[server]",
            "desc": "Get the list of public rooms" " on the specified server.",
            "shortdesc": "List the rooms.",
            "completion": completions.list,
        },
        {
            "name": "message",
            "func": commands.message,
            "usage": "<jid> [optional message]",
            "desc": (
                "Open a conversation with the specified JID (even if it"
                " is not in our roster), and send a message to it, if the "
                "message is specified."
            ),
            "shortdesc": "Send a message",
            "completion": completions.message,
        },
        {
            "name": "version",
            "func": commands.version,
            "usage": "<jid>",
            "desc": (
                "Get the software version of the given JID (usually its"
                " XMPP client and Operating System)."
            ),
            "shortdesc": "Get the software version of a JID.",
            "completion": completions.version,
        },
        {
            "name": "server_cycle",
            "func": commands.server_cycle,
            "usage": "[domain] [message]",
            "desc": "Disconnect and reconnect in all the rooms in domain.",
            "shortdesc": "Cycle a range of rooms",
            "completion": completions.server_cycle,
        },
        {
            "name": "bind",
            "func": commands.bind,
            "usage": "<key> <equ>",
            "desc": (
                "Bind a key to another key or to a “command”. For "
                'example "/bind ^H KEY_UP" makes Control + h do the'
                " same same as the Up key."
            ),
            "completion": completions.bind,
            "shortdesc": "Bind a key to another key.",
        },
        {
            "name": "load",
            "func": commands.load,
            "usage": "<plugin> [<otherplugin> …]",
            "shortdesc": "Load the specified plugin(s)",
            "completion": plugin_manager.completion_load,
        },
        {
            "name": "unload",
            "func": commands.unload,
            "usage": "<plugin> [<otherplugin> …]",
            "shortdesc": "Unload the specified plugin(s)",
            "completion": plugin_manager.completion_unload,
        },
        {
            "name": "plugins",
            "func": commands.plugins,
            "shortdesc": "Show the plugins in use.",
        },
        {
            "name": "presence",
            "func": commands.presence,
            "usage": "<JID> [type] [status]",
            "desc": "Send a directed presence to <JID> and using"
            " [type] and [status] if provided.",
            "shortdesc": "Send a directed presence.",
            "completion": completions.presence,
        },
        {
            "name": "rawxml",
            "func": commands.rawxml,
            "usage": "<xml>",
            "shortdesc": "Send a custom xml stanza.",
        },
        {
            "name": "invite",
            "func": commands.invite,
            "usage": "<jid> <room> [reason]",
            "desc": "Invite jid in room with reason.",
            "shortdesc": "Invite someone in a room.",
            "completion": completions.invite,
        },
        {
            "name": "impromptu",
            "func": commands.impromptu,
            "usage": "<jid> [jid ...]",
            "desc": "Invite specified JIDs into a newly created room.",
            "shortdesc": "Invite specified JIDs into newly created room.",
            "completion": completions.impromptu,
        },
        {
            "name": "invitations",
            "func": commands.invitations,
            "shortdesc": "Show the pending invitations.",
        },
        {
            "name": "bookmarks",
            "func": commands.bookmarks,
            "shortdesc": "Show the current bookmarks.",
        },
        {
            "name": "remove_bookmark",
            "func": commands.remove_bookmark,
            "usage": "[jid]",
            "desc": "Remove the specified bookmark, or the "
            "bookmark on the current tab, if any.",
            "shortdesc": "Remove a bookmark",
            "completion": completions.remove_bookmark,
        },
        {
            "name": "xml_tab",
            "func": commands.xml_tab,
            "shortdesc": "Open an XML tab.",
        },
        {
            "name": "runkey",
            "func": commands.runkey,
            "usage": "<key>",
            "shortdesc": "Execute the action defined for <key>.",
            "completion": completions.runkey,
        },
        {
            "name": "self",
            "func": commands.self_,
            "shortdesc": "Remind you of who you are.",
        },
        {
            "name": "last_activity",
            "func": commands.last_activity,
            "usage": "<jid>",
            "desc": "Informs you of the last activity of a JID.",
            "shortdesc": "Get the activity of someone.",
            "completion": completions.last_activity,
        },
        {
            "name": "ad-hoc",
            "func": commands.adhoc,
            "usage": "<jid>",
            "shortdesc": "List available ad-hoc commands on the given jid",
        },
        {
            "name": "reload",
            "func": commands.reload,
            "shortdesc": "Reload the config. You can achieve the same by "
            "sending SIGUSR1 to poezio.",
        },
        {
            "name": "debug",
            "func": commands.debug,
            "usage": "[debug_filename]",
            "shortdesc": "Enable or disable debug logging according to the "
            "presence of [debug_filename].",
        },
    ]



"""
Module defining structures useful to the core class and related methods
"""
from __future__ import annotations
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    TYPE_CHECKING,
    Optional,
    TypedDict,
    Concatenate,
    ParamSpec,
    TypeVar,
)

if TYPE_CHECKING:
    from poezio import windows

P = ParamSpec('P')
T = TypeVar('T')

__all__ = [
    'Command',
    'Completion',
    'POSSIBLE_SHOW',
    'Status',
]

POSSIBLE_SHOW = {
    'available': None,
    'chat': 'chat',
    'away': 'away',
    'afk': 'away',
    'dnd': 'dnd',
    'busy': 'dnd',
    'xa': 'xa'
}


@dataclass(slots=True)
class Status:
    show: str
    message: str


class Completion:
    """
    A completion result essentially currying the input completion call.
    """
    __slots__ = ('func', 'args', 'kwargs', 'comp_list')

    def __init__(
        self,
        func: Callable[Concatenate[list[str], P], bool],
        comp_list: list[str],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        self.func = func
        self.comp_list = comp_list
        self.args = args
        self.kwargs = kwargs

    def run(self) -> bool:
        return self.func(self.comp_list, *self.args, **self.kwargs)


GenericCompletion = (
    Callable[['windows.Input'], Completion | None]
    | Callable[['windows.Input'],  Completion]
    | None
)


@dataclass(slots=True)
class Command:
    func: Callable[..., Any]
    desc: str
    comp: Optional[Callable[['windows.Input'], Completion | None]]
    short_desc: str
    usage: str


class CommandKwargs(TypedDict, total=False):
    desc: str
    shortdesc: str
    completion: Callable[['windows.Input'], Completion | None] | None
    usage: str



# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
Defines the Resource and Contact classes, which are used in
the roster.
"""

from collections import defaultdict
import logging
from typing import (
    Any,
    Dict,
    Iterator,
    List,
)

from slixmpp import InvalidJID, JID
from slixmpp.roster import RosterItem
from slixmpp.types import ResourceDict

log = logging.getLogger(__name__)


class Resource:
    """
    Defines a roster item.
    It's a precise resource.
    """

    def __init__(self, jid: str, data: ResourceDict):
        """
        data: the dict to use as a source
        """
        # Full JID
        self._jid: str = jid
        self._data: ResourceDict = data

    @property
    def jid(self) -> str:
        """JID of this resource."""
        return self._jid

    @property
    def priority(self) -> int:
        """Value of the priority received for this resource."""
        try:
            return int(self._data.get('priority', 0))
        except Exception:
            return 0

    @property
    def presence(self) -> str:
        """Value of the latest "show" for this resource."""
        return str(self._data.get('show')) or ''

    @property
    def status(self) -> str:
        """Value of the latest status for this resource."""
        return str(self._data.get('status')) or ''

    def __repr__(self) -> str:
        return '<%s>' % self._jid

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Resource):
            return False
        return self.jid == value.jid and self._data == value._data


class Contact:
    """
    This a way to gather multiple resources from the same bare JID.
    This class contains zero or more Resource object and useful methods
    to get the resource with the highest priority, etc
    """

    def __init__(self, item: RosterItem):
        """
        item: a slixmpp RosterItem pointing to that contact
        """
        self.__item = item
        self.folded_states: Dict[str, bool] = defaultdict(lambda: True)
        self._name = ''
        self.avatar = None
        self.error = None
        self.rich_presence: Dict[str, Any] = defaultdict(lambda: None)

    @property
    def groups(self) -> List[str]:
        """Name of the groups the contact is in"""
        return self.__item['groups'] or ['none']

    @property
    def bare_jid(self) -> JID:
        """The bare jid of the contact"""
        return JID(self.__item.jid)

    @property
    def name(self) -> str:
        """The name of the contact or an empty string."""
        return self.__item['name'] or self._name or ''

    @name.setter
    def name(self, value:  str) -> None:
        """Set the name of the contact with user nickname"""
        self._name = value

    @property
    def ask(self) -> str | None:
        """If we asked the contact for presence or not."""
        if self.__item['pending_out']:
            return 'asked'
        return None

    @property
    def pending_in(self) -> bool:
        """We received a subscribe stanza from this contact."""
        return self.__item['pending_in']

    @pending_in.setter
    def pending_in(self, value: bool) -> None:
        self.__item['pending_in'] = value

    @property
    def pending_out(self) -> bool:
        """We sent a subscribe stanza to this contact."""
        return self.__item['pending_out']

    @pending_out.setter
    def pending_out(self, value: bool) -> None:
        self.__item['pending_out'] = value

    @property
    def resources(self) -> Iterator[Resource]:
        """List of the available resources as Resource objects"""
        return (Resource('%s%s' % (self.bare_jid, ('/' + key)
                                   if key else ''), self.__item.resources[key])
                for key in self.__item.resources.keys())

    @property
    def subscription(self) -> str:
        """Status of our subscription to this contact."""
        return self.__item['subscription']

    def __contains__(self, value:  str | JID) -> bool:
        try:
            resource = JID(value).resource
        except InvalidJID:
            resource = None
        return value in self.__item.resources or \
            (resource is not None and resource in self.__item.resources)

    def __len__(self) -> int:
        """Number of resources"""
        return len(self.__item.resources)

    def __bool__(self) -> bool:
        """This contact exists even when he has no resources"""
        return True

    def __getitem__(self, key: str | JID) -> Resource | None:
        """Return the corresponding Resource object, or None"""
        try:
            res = JID(key).resource
        except InvalidJID:
            return None
        resources = self.__item.resources
        item = resources.get(res, None) or resources.get(str(key), None)
        return Resource(str(key), item) if item else None

    def subscribe(self) -> None:
        """Subscribe to this JID"""
        self.__item.subscribe()

    def authorize(self) -> None:
        """Authorize this JID"""
        self.__item.authorize()

    def unauthorize(self) -> None:
        """Unauthorize this JID"""
        self.__item.unauthorize()

    def unsubscribe(self) -> None:
        """Unsubscribe from this JID"""
        self.__item.unsubscribe()

    def get(self, key: str,
            default: Resource | None = None) -> Resource | None:
        """Same as __getitem__, but with a configurable default"""
        return self[key] or default

    def get_resources(self) -> list[Resource]:
        """Return all resources, sorted by priority """
        def compare_resources(res: Resource) -> int:
            return res.priority
        return sorted(self.resources, key=compare_resources, reverse=True)

    def get_highest_priority_resource(self) -> Resource | None:
        """Return the resource with the highest priority"""
        resources = self.get_resources()
        if resources:
            return resources[0]
        return None

    def folded(self, group_name: str = 'none') -> bool:
        """
        Return the Folded state of a contact for this group
        """
        return self.folded_states[group_name]

    def toggle_folded(self, group: str = 'none') -> None:
        """
        Fold if it's unfolded, and vice versa
        """
        self.folded_states[group] = not self.folded_states[group]

    def __repr__(self) -> str:
        ret = '<Contact: %s' % self.bare_jid
        for resource in self.resources:
            ret += '\n\t\t%s' % resource
        return ret + ' />\n'



import pytest
from contextlib import contextmanager
from datetime import datetime
from poezio.theming import get_theme
from poezio.ui.render import build_lines, write_pre, ReactionLine
from poezio.ui.types import BaseMessage, Message, StatusMessage, XMLLog
from poezio.xhtml import clean_text


def test_simple_build_basemsg():
    msg = BaseMessage(txt='coucou')
    line = build_lines(msg, 100, True, 10)[0]
    assert (line.start_pos, line.end_pos) == (0, 6)


def test_simple_render_message():
    msg = Message(txt='coucou', nickname='toto')
    line = build_lines(msg, 100, True, 10)[0]
    assert (line.start_pos, line.end_pos) == (0, 6)


def test_simple_render_xmllog():
    msg = XMLLog(txt='coucou', incoming=True)
    line = build_lines(msg, 100, True, 10)[0]
    assert (line.start_pos, line.end_pos) == (0, 6)


def test_simple_render_separator():
    line = build_lines(None, 100, True, 10)[0]
    assert line is None


def test_render_reactions():
    msg = Message(
        txt="coucou",
        nickname="test",
        reactions={'😭': {'a', 'b'}, '🫠': {'1', '2', '3', '4'}, '🎉': {'x'}},
    )
    lines = build_lines(msg, 100, True, 10)
    line = lines[0]
    assert (line.start_pos, line.end_pos) == (0, 6)

    reaction_line = lines[1]
    assert isinstance(reaction_line, ReactionLine)
    assert reaction_line.txt == '\x195,-1}| 🫠\x195,-1}[4]\x19o 😭\x195,-1}[2]\x19o 🎉'
    assert clean_text(reaction_line.txt) == '| 🫠[4] 😭[2] 🎉'


def test_simple_render_status():
    class Obj:
        name = 'toto'
    msg = StatusMessage("Coucou {name}", {'name': lambda: Obj.name})
    assert msg.txt == "Coucou toto"
    Obj.name = 'titi'
    build_lines(msg, 100, True, 10)[0]
    assert msg.txt == "Coucou titi"


class FakeBuffer:
    def __init__(self):
        self.text = ''

    @contextmanager
    def colored_text(self, *args, **kwargs):
        yield None

    def addstr(self, txt):
        self.text += txt


@pytest.fixture(scope='function')
def buffer():
    return FakeBuffer()


@pytest.fixture
def time():
    return datetime.strptime('2019-09-27 10:11:12', '%Y-%m-%d %H:%M:%S')


def test_write_pre_basemsg(buffer):
    str_time = '10:11:12'
    time = datetime.strptime(str_time, '%H:%M:%S')
    msg = BaseMessage(txt='coucou', time=time)
    size = write_pre(msg, buffer, True, 10)
    assert buffer.text == '10:11:12 '
    assert size == len(buffer.text)


def test_write_pre_message_simple(buffer, time):
    msg = Message(txt='coucou', nickname='toto', time=time)
    size = write_pre(msg, buffer, True, 10)
    assert buffer.text == '10:11:12 toto> '
    assert size == len(buffer.text)


def test_write_pre_message_simple_history(buffer, time):
    msg = Message(txt='coucou', nickname='toto', time=time, history=True)
    size = write_pre(msg, buffer, True, 10)
    assert buffer.text == '2019-09-27 10:11:12 toto> '
    assert size == len(buffer.text)


def test_write_pre_message_highlight(buffer, time):
    msg = Message(txt='coucou', nickname='toto', time=time, highlight=True)
    size = write_pre(msg, buffer, True, 10)
    assert buffer.text == '10:11:12 toto> '
    assert size == len(buffer.text)


def test_write_pre_message_no_timestamp(buffer):
    msg = Message(txt='coucou', nickname='toto')
    size = write_pre(msg, buffer, False, 10)
    assert buffer.text == 'toto> '
    assert size == len(buffer.text)


def test_write_pre_message_me(buffer, time):
    msg = Message(txt='/me coucou', nickname='toto', time=time)
    size = write_pre(msg, buffer, True, 10)
    assert buffer.text == '10:11:12 * toto '
    assert size == len(buffer.text)


def test_write_pre_message_revisions(buffer, time):
    msg = Message(txt='coucou', nickname='toto', time=time, revisions=5)
    size = write_pre(msg, buffer, True, 10)
    assert buffer.text == '10:11:12 toto5> '
    assert size == len(buffer.text)


def test_write_pre_message_revisions_me(buffer, time):
    msg = Message(txt='/me coucou', nickname='toto', time=time, revisions=5)
    size = write_pre(msg, buffer, True, 10)
    assert buffer.text == '10:11:12 * toto5 '
    assert size == len(buffer.text)


def test_write_pre_message_ack(buffer, time):
    ack = get_theme().CHAR_ACK_RECEIVED
    expected = '10:11:12 %s toto> ' % ack
    msg = Message(txt='coucou', nickname='toto', time=time, ack=1)
    size = write_pre(msg, buffer, True, 10)
    assert buffer.text == expected
    assert size == len(buffer.text)


def test_write_pre_message_nack(buffer, time):
    nack = get_theme().CHAR_NACK
    expected = '10:11:12 %s toto> ' % nack
    msg = Message(txt='coucou', nickname='toto', time=time, ack=-1)
    size = write_pre(msg, buffer, True, 10)
    assert buffer.text == expected
    assert size == len(buffer.text)


def test_write_pre_xmllog_in(buffer):
    msg = XMLLog(txt="coucou", incoming=True)
    size = write_pre(msg, buffer, True, 10)
    assert buffer.text == '%s IN  ' % msg.time.strftime('%H:%M:%S')
    assert size == len(buffer.text)


def test_write_pre_xmllog_out(buffer):
    msg = XMLLog(txt="coucou", incoming=False)
    size = write_pre(msg, buffer, True, 10)
    assert buffer.text == '%s OUT ' % msg.time.strftime('%H:%M:%S')
    assert size == len(buffer.text)



"""
UntrackMe wannabe plugin.

This plugin replaces instances of terrible and closed services with their
FOSS frontends when available, as well as remove the usual tracking parameters
from URLs.

Config options
--------------

.. code-block:: ini

    [untrackme]
    # Cleanup the tracking parameters from the URLs
    cleanup = true
    # Redirect bad services to better frontends
    redirect = true
    # Log corrections to modified messages
    display_corrections = false

    [services]
    invidious = https://invidious.snopyta.org
    nitter = https://xcancel.com
    bibliogram =  https://bibliogram.art


Three redirect services exist:

* Invidious (for youtube)
* Nitter (for twitter)
* Bibliogram (for instagram)

But bibliogram is abandoned.

"""

from typing import Callable, Dict, Tuple, Union

import re
import logging
from slixmpp import Message
from poezio import tabs
from poezio.plugin import BasePlugin
from urllib.parse import quote as urlquote


log = logging.getLogger(__name__)

ChatTabs = Union[
    tabs.MucTab,
    tabs.DynamicConversationTab,
    tabs.StaticConversationTab,
    tabs.PrivateTab,
]

RE_URL: re.Pattern[str] = re.compile('https?://(?P<host>[^/]+)(?P<rest>[^ ]*)')

SERVICES: Dict[str, Tuple[str, bool]] = {  # host: (service, proxy)
    'm.youtube.com': ('invidious', False),
    'www.youtube.com': ('invidious', False),
    'youtube.com': ('invidious', False),
    'youtu.be': ('invidious', False),
    'youtube-nocookie.com': ('invidious', False),
    'mobile.twitter.com': ('nitter', False),
    'www.twitter.com': ('nitter', False),
    'twitter.com': ('nitter', False),
    'x.com': ('nitter', False),
    'pic.twitter.com': ('nitter_img', True),
    'pbs.twimg.com': ('nitter_img', True),
    'instagram.com': ('bibliogram', False),
    'www.instagram.com': ('bibliogram', False),
    'm.instagram.com': ('bibliogram', False),
}


def proxy(service: str) -> Callable[[str], str]:
    """Some services require the original url"""
    def inner(origin: str) -> str:
        return service + urlquote(origin)
    return inner


class Plugin(BasePlugin):
    """UntrackMe"""

    default_config: Dict[str, Dict[str, Union[str, bool]]] = {
        'default': {
            'cleanup': True,
            'redirect': True,
            'display_corrections': False,
        },
        'services': {
            'invidious': 'https://invidious.snopyta.org',
            'nitter': 'https://xcancel.com',
            'bibliogram': 'https://bibliogram.art',
        },
    }

    def init(self) -> None:
        nitter_img = self.config.get('nitter', section='services') + '/pic/'
        self.config.set('nitter_img', nitter_img, section='services')

        self.api.add_event_handler('muc_say', self.handle_msg)
        self.api.add_event_handler('conversation_say', self.handle_msg)
        self.api.add_event_handler('private_say', self.handle_msg)

        self.api.add_event_handler('muc_msg', self.handle_msg)
        self.api.add_event_handler('conversation_msg', self.handle_msg)
        self.api.add_event_handler('private_msg', self.handle_msg)

    def map_services(self, match: re.Match[str]) -> str:
        """
            If it matches a host that we know about, change the domain for the
            alternative service. Some hosts needs to be proxied instead (such
            as twitter pictures), so they're url encoded and appended to the
            proxy service.
        """

        host = match.group('host')

        result: str
        dest = SERVICES.get(host)
        if dest is None:
            return match.group(0)

        destname, proxy = dest
        replaced = self.config.get(destname, section='services')
        result = replaced + match.group('rest')

        if proxy:
            url = urlquote(match.group(0))
            result = replaced + url

            # TODO: count parenthesis?
            # Removes comma at the end of a link.
            if result[-3] == '%2C':
                result = result[:-3] + ','

        return result

    def handle_msg(self, msg: Message, tab: ChatTabs) -> None:
        orig = msg['body']

        if self.config.get('cleanup', section='default'):
            msg['body'] = self.cleanup_url(msg['body'])
        if self.config.get('redirect', section='default'):
            msg['body'] = self.redirect_url(msg['body'])

        if self.config.get('display_corrections', section='default') and \
           msg['body'] != orig:
            log.debug(
                'UntrackMe in tab \'%s\':\nOriginal: %s\nModified: %s',
                tab.name, orig, msg['body'],
            )

            self.api.information(
                'UntrackMe in tab \'{}\':\nOriginal: {}\nModified: {}'.format(
                    tab.name, orig, msg['body']
                ),
                'Info',
            )

    def cleanup_url(self, txt: str) -> str:
        # fbclid: used globally (Facebook)
        # utm_*: used globally https://en.wikipedia.org/wiki/UTM_parameters
        # ncid: DoubleClick (Google)
        # ref_src, ref_url: twitter
        # Others exist but are excluded because they are not common.
        # See https://en.wikipedia.org/wiki/UTM_parameters
        return re.sub(
            (
                '(https?://[^ ]+)&?'
                '(fbclid|dclid|ncid|utm_source|utm_medium|utm_campaign|utm_term|utm_content|ref_src|ref_url)'
                '=[^ &#]*'
            ),
            r'\1',
            txt,
        )

    def redirect_url(self, txt: str) -> str:
        return RE_URL.sub(self.map_services, txt)



"""
Defines the data-forms Tab
"""

import logging
from typing import Any, Callable, TYPE_CHECKING, Concatenate, ParamSpec

from poezio import windows
from poezio.tabs import Tab, NoInputTab
from poezio.core.structs import Command
from slixmpp.plugins.xep_0004 import Form

if TYPE_CHECKING:
    from poezio.core.core import Core

log = logging.getLogger(__name__)

P = ParamSpec('P')


class DataFormsTab(NoInputTab):
    """
    A tab containing various window type, displaying
    a form that the user needs to fill.
    """
    plugin_commands: dict[str, Command] = {}
    plugin_keys: dict[str, Callable[..., Any]] = {}

    def __init__(self, core: 'Core', form: Form, on_cancel: Callable[Concatenate[Form, P], None], on_send: Callable[Concatenate[Form, P], None], *args: P.args, **kwargs: P.kwargs) -> None:
        NoInputTab.__init__(self, core)
        self._form = form
        self._on_cancel = on_cancel
        self._on_send = on_send
        self._args = args
        self._kwargs = kwargs
        self.fields = []
        for field in self._form:
            self.fields.append(field)
        self.topic_win = windows.Topic()
        self.form_win = windows.FormWin(form, self.height - 4, self.width, 1,
                                        0)
        self.help_win = windows.HelpText("Ctrl+Y: send form, Ctrl+G: cancel")
        self.help_win_dyn = windows.HelpText()
        self.key_func['KEY_UP'] = self.form_win.go_to_previous_input
        self.key_func['KEY_DOWN'] = self.form_win.go_to_next_input
        self.key_func['^G'] = self.on_cancel
        self.key_func['^Y'] = self.on_send
        self.resize()
        self.update_commands()

    def on_cancel(self) -> bool:
        self._on_cancel(self._form, *self._args, **self._kwargs)
        return True

    def on_send(self) -> bool:
        self._form.reply()
        self.form_win.reply()
        self._on_send(self._form, *self._args, **self._kwargs)
        return True

    def on_input(self, key: str, raw: bool = False) -> None:
        if key in self.key_func:
            res = self.key_func[key]()
            if res:
                return
            self.help_win_dyn.refresh_txt(self.form_win.get_help_message())
            self.form_win.refresh_current_input()
        else:
            self.form_win.on_input(key, raw=raw)

    def resize(self) -> None:
        self.need_resize = False
        self.topic_win.resize(1, self.width, 0, 0)
        self.form_win.resize(self.height - 3 - Tab.tab_win_height(),
                             self.width, 1, 0)
        self.help_win.resize(1, self.width, self.height - 1, 0)
        self.help_win_dyn.resize(1, self.width,
                                 self.height - 2 - Tab.tab_win_height(), 0)

    def refresh(self) -> None:
        if self.need_resize:
            self.resize()
        self.topic_win.refresh(self._form['title'])
        self.refresh_tab_win()
        self.help_win.refresh()
        self.help_win_dyn.refresh_txt(self.form_win.get_help_message())
        self.form_win.refresh()



# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
Defines the Roster and RosterGroup classes
"""
import logging

from poezio.config import config
from poezio.contact import Contact
from poezio.roster_sorting import SORTING_METHODS, GROUP_SORTING_METHODS

from os import path as p
from collections.abc import Iterator
from datetime import datetime
from slixmpp.exceptions import IqError, IqTimeout
from slixmpp.roster.single import RosterNode
from slixmpp import JID, InvalidJID
from typing import Callable

log = logging.getLogger(__name__)


class Roster:
    """
    The proxy class to get the roster from slixmpp.
    Caches Contact and RosterGroup objects.
    """

    DEFAULT_FILTER = (lambda _, __: True, '')

    __node: RosterNode
    length: int
    connected: int
    folded_groups: set[str]
    last_built: datetime
    last_modified: datetime
    groups: dict[str, 'RosterGroup']
    contacts: dict[str | JID, Contact]
    contact_filter: tuple[
        Callable[[Contact, str], bool],
        str,
    ]

    def __init__(self) -> None:
        self.__node = RosterNode(None, None)  # type: ignore

        # A tuple(function, *args) function to filter contacts
        # on search, for example
        self.contact_filter = self.DEFAULT_FILTER
        self.groups = {}
        self.contacts = {}
        self.length = 0
        self.connected = 0
        self.folded_groups = set()

        # Used for caching roster infos
        self.last_built = datetime.now()
        self.last_modified = datetime.now()

    def reset(self) -> None:
        """
        node: the RosterSingle from slixmpp
        """
        self.__node = RosterNode(None, None)  # type: ignore

        # A tuple(function, *args) function to filter contacts
        # on search, for example
        self.contact_filter = self.DEFAULT_FILTER
        self.folded_groups = set(
            config.getlist('folded_roster_groups', section='var')
        )
        self.groups = {}
        self.contacts = {}
        self.length = 0
        self.connected = 0

        # Used for caching roster infos
        self.last_built = datetime.now()
        self.last_modified = datetime.now()

    def modified(self) -> None:
        self.last_modified = datetime.now()

    @property
    def needs_rebuild(self) -> bool:
        return self.last_modified >= self.last_built

    def __getitem__(self, key: str | JID) -> Contact | None:
        """Get a Contact from his bare JID"""
        try:
            key = JID(key).bare
        except InvalidJID:
            return None
        if key in self.contacts and self.contacts[key] is not None:
            return self.contacts[key]
        if key in self.jids():
            contact = Contact(self.__node[key])
            self.contacts[key] = contact
            return contact
        return None

    def __setitem__(self, key: str | JID, value: Contact) -> None:
        """Set the a Contact value for the bare jid key"""
        self.contacts[key] = value

    def remove(self, jid: str | JID) -> None:
        """Send a removal iq to the server"""
        try:
            jid = JID(jid).bare
        except InvalidJID:
            return
        if self.__node[jid]:
            try:
                self.__node[jid].send_presence(ptype='unavailable')
                self.__node.remove(jid)
            except (IqError, IqTimeout):
                log.debug('IqError when removing %s:', jid, exc_info=True)

    def __delitem__(self, jid: str | JID) -> None:
        """Remove a contact from the roster view"""
        try:
            jid = JID(jid).bare
        except InvalidJID:
            return
        contact = self[jid]
        if not contact:
            return
        del self.contacts[contact.bare_jid]

        for group in list(self.groups.values()):
            group.remove(contact)
            if not group:
                del self.groups[group.name]
        self.modified()

    def __iter__(self) -> Iterator[Contact]:
        """Iterate over the jids of the contacts"""
        return iter(self.contacts.values())

    def __contains__(self, key: str | JID) -> bool:
        """True if the bare jid is in the roster, false otherwise"""
        try:
            return JID(key).bare in self.jids()
        except InvalidJID:
            return False

    @property
    def jid(self) -> JID | str:
        """Our JID"""
        return self.__node.jid

    def get_and_set(self, jid: str | JID) -> Contact:
        contact = self.contacts.get(jid)
        if contact is None:
            contact = Contact(self.__node[jid])
            self.contacts[jid] = contact
            return contact
        return contact

    def set_node(self, value: RosterNode) -> None:
        """Set the slixmpp RosterSingle for our roster"""
        self.__node = value

    def get_groups(self, sort: str = '') -> list['RosterGroup']:
        """Return a list of the RosterGroups"""
        group_list = sorted(
            (group for group in self.groups.values() if group),
            key=lambda x: x.name.lower() if x.name else '')

        for sorting in sort.split(':'):
            if sorting == 'reverse':
                group_list = list(reversed(group_list))
            else:
                method = GROUP_SORTING_METHODS.get(sorting, lambda _: 0)
                group_list = sorted(group_list, key=method)
        return group_list

    def get_group(self, name: str) -> 'RosterGroup':
        """Return a group or create it if not present"""
        if name in self.groups:
            return self.groups[name]
        group = RosterGroup(
            name, folded=name in self.folded_groups
        )
        self.groups[name] = group
        return group

    def add(self, jid: str | JID) -> None:
        """Subscribe to a jid"""
        self.__node.subscribe(jid)

    def jids(self) -> list[JID | str]:
        """List of the contact JIDs"""
        jids: list[str | JID] = []
        for key in self.__node.keys():
            contact = self.get_and_set(key)
            if key and key != self.jid and (contact and self.exists(contact)):
                jids.append(key)
        self.update_size(jids)
        return jids

    def update_size(self, jids: list[JID | str] | None = None) -> None:
        if jids is None:
            jids = self.jids()
        self.length = len(jids)

    def get_contacts(self) -> list[Contact]:
        """
        Return a list of all the contacts
        """
        return [self.contacts[jid] for jid in self.jids()]

    def get_contacts_sorted_filtered(self, sort: str = '') -> list[Contact]:
        """
        Return a list of all the contacts sorted with a criteria
        """
        contact_list = []
        for contact in self.get_contacts():
            if contact.bare_jid != self.jid:
                if self.contact_filter is not self.DEFAULT_FILTER:
                    if self.contact_filter[0](contact, self.contact_filter[1]):
                        contact_list.append(contact)
                else:
                    contact_list.append(contact)
        contact_list = sorted(contact_list, key=SORTING_METHODS['name'])

        for sorting in sort.split(':'):
            if sorting == 'reverse':
                contact_list = list(reversed(contact_list))
            else:
                method = SORTING_METHODS.get(sorting, lambda _: 0)
                contact_list = sorted(contact_list, key=method)
        return contact_list

    def save_to_config_file(self) -> bool:
        """
        Save various information to the config file
        e.g. the folded groups
        """
        folded_groups = ':'.join([
            group.name for group in self.groups.values()
            if group.folded
        ])
        log.debug('folded:%s\n', folded_groups)
        return config.silent_set('folded_roster_groups', folded_groups, 'var')

    def get_nb_connected_contacts(self) -> int:
        """
        Get the number of connected contacts
        """
        return self.connected

    def update_contact_groups(self, contact: Contact | str) -> None:
        """Regenerate the RosterGroups when receiving a contact update"""
        if not isinstance(contact, Contact):
            contact = self.get_and_set(contact)
        if not contact:
            return
        for name, group in self.groups.items():
            if name in contact.groups and contact not in group:
                group.add(contact)
            elif contact in group and name not in contact.groups:
                group.remove(contact)

        for name in contact.groups:
            if name not in self.groups:
                self.groups[name] = RosterGroup(
                    name, folded=name in self.folded_groups)
                self.groups[name].add(contact)

    def __len__(self) -> int:
        """
        Return the number of contacts
        (used to return the display size, but now we have
        the display cache in RosterWin for that)
        """
        return self.length

    def __repr__(self) -> str:
        ret = '== Roster:\nContacts:\n'
        for contact in self.contacts.values():
            ret += '%s\n' % (contact, )
        ret += 'Groups\n'
        for group in self.groups:
            ret += '%s\n' % (group, )
        return ret + '\n'

    def export(self, path: str) -> bool:
        """Export a list of bare jids to a given file"""
        if p.isfile(path):
            return False
        try:
            f = open(path, 'w+', encoding='utf-8')
            f.writelines([
                str(i) + "\n" for i in self.contacts
                if self.contacts[i] and (self.contacts[i].subscription == "both" or self.contacts[i].ask)
            ])
            f.close()
        except OSError:
            return False
        return True

    def exists(self, contact: Contact | None) -> bool:
        if not contact:
            return False
        for group in contact.groups:
            if contact not in self.groups.get(group, tuple()):
                return False
        return True


class RosterGroup:
    """
    A RosterGroup is a group containing contacts
    It can be Friends/Family etc, but also can be
    Online/Offline or whatever
    """

    def __init__(self, name: str | None,
                 contacts: list[Contact] | None = None,
                 folded: bool = False) -> None:
        if not contacts:
            contacts = []
        self.contacts = set(contacts)
        self.name = name if name is not None else ''
        self.folded = folded  # if the group content is to be shown

    def __iter__(self) -> Iterator[Contact]:
        """Iterate over the contacts"""
        return iter(self.contacts)

    def __repr__(self) -> str:
        return '<Roster_group: %s; %s>' % (self.name, self.contacts)

    def __len__(self) -> int:
        """Number of contacts in the group"""
        return len(self.contacts)

    def __contains__(self, contact: Contact) -> bool:
        """
        Return a bool, telling if the contact is in the group
        """
        return contact in self.contacts

    def add(self, contact: Contact) -> None:
        """Add a contact to the group"""
        self.contacts.add(contact)

    def remove(self, contact: Contact) -> None:
        """Remove a contact from the group if present"""
        if contact in self.contacts:
            self.contacts.remove(contact)

    def get_contacts(self,
                     contact_filter: tuple[
                         Callable[[Contact, str], bool], str
                     ] | None = None,
                     sort: str = '') -> list[Contact]:
        """Return the group contacts, filtered and sorted"""
        if contact_filter is Roster.DEFAULT_FILTER or contact_filter is None:
            contact_list = list(self.contacts)
        else:
            contact_list = [
                contact for contact in self.contacts
                if contact_filter[0](contact, contact_filter[1])
            ]
        contact_list = sorted(contact_list, key=SORTING_METHODS['name'])

        for sorting in sort.split(':'):
            if sorting == 'reverse':
                contact_list = list(reversed(contact_list))
            else:
                method = SORTING_METHODS.get(sorting, lambda _: 0)
                contact_list = sorted(contact_list, key=method)
        return contact_list

    def toggle_folded(self) -> None:
        """Fold/unfold the group in the roster"""
        self.folded = not self.folded
        if self.folded:
            if self.name not in roster.folded_groups:
                roster.folded_groups.add(self.name)
        else:
            if self.name in roster.folded_groups:
                roster.folded_groups.remove(self.name)

    def get_nb_connected_contacts(self) -> int:
        """Return the number of connected contacts"""
        return len([1 for contact in self.contacts if len(contact)])


# Shared roster object
roster = Roster()



import pytest
from poezio.windows import Input, HistoryInput, MessageInput
from poezio import config


class ConfigShim:
    def get(self, *args, **kwargs):
        return ''

    def getbool(self, *args, **kwargs):
        return True


config.config = ConfigShim()


class SubInput(Input):

    def rewrite_text(self, *args, **kwargs):
        return None


@pytest.fixture
def input():
    from poezio.windows import base_wins
    base_wins.TAB_WIN = True  # The value is not relevant
    return SubInput()


class TestInput(object):

    def test_do_command(self, input):

        input.do_command('a')
        assert input.text == 'a'

        for char in 'coucou':
            input.do_command(char)
        assert input.text == 'acoucou'

    def test_empty(self, input):
        assert input.is_empty()
        input.do_command('a')
        assert not input.is_empty()

    def test_key_left(self, input):
        for char in 'this is a line':
            input.do_command(char)
        for i in range(4):
            input.key_left()
        for char in 'long ':
            input.do_command(char)

        assert input.text == 'this is a long line'

    def test_key_right(self, input):
        for char in 'this is a line':
            input.do_command(char)
        for i in range(4):
            input.key_left()
        input.key_right()

        for char in 'iii':
            input.do_command(char)

        assert input.text == 'this is a liiiine'

    def test_key_home(self, input):
        for char in 'this is a line of text':
            input.do_command(char)
        input.do_command('z')
        input.key_home()
        input.do_command('a')

        assert input.text == 'athis is a line of textz'

    def test_key_end(self, input):
        for char in 'this is a line of text':
            input.do_command(char)
        input.key_home()
        input.key_end()
        input.do_command('z')

        assert input.text == 'this is a line of textz'

